"""An implementation of an integer string based search space."""
from typing import Final

import numpy

from moptipy.spaces.nparrayspace import NPArraySpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.types import type_error

#: the minimum value
KEY_MIN: Final[str] = "min"
#: the maximum value
KEY_MAX: Final[str] = "max"


class IntSpace(NPArraySpace):
    """
    A space where each element is a one-dimensional numpy integer array.

    Such spaces can serve as basis for implementing combinatorial
    optimization and can be extended to host permutations.
    """

    def __init__(self, dimension: int,
                 min_value: int,
                 max_value: int) -> None:
        """
        Create the integer-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param min_value: the minimum value
        :param max_value: the maximum value
        """
        super().__init__(
            dimension,
            int_range_to_dtype(min_value=min_value, max_value=max_value))
        #: The minimum permitted value.
        self.min_value: Final[int] = min_value
        #: The maximum permitted value.
        self.max_value: Final[int] = max_value

    def create(self) -> numpy.ndarray:
        """
        Create an integer vector filled with the minimal value.

        :return: the vector

        >>> from moptipy.spaces.intspace import IntSpace
        >>> s = IntSpace(dimension=12, min_value=5, max_value=332)
        >>> v = s.create()
        >>> print(s.to_str(v))
        5,5,5,5,5,5,5,5,5,5,5,5
        >>> print(v.dtype)
        int16
        """
        return numpy.full(shape=self.dimension,
                          fill_value=self.min_value,
                          dtype=self.dtype)

    def from_str(self, text: str) -> numpy.ndarray:
        """
        Convert a string to an integer string.

        :param text: the text
        :return: the vector
        :raises TypeError: if `text` is not a `str`
        :raises ValueError: if `text` cannot be converted to a valid vector
        """
        if not (isinstance(text, str)):
            raise type_error(text, "text", str)
        x = numpy.fromstring(text, dtype=self.dtype, sep=",")
        self.validate(x)
        return x

    def validate(self, x: numpy.ndarray) -> None:
        """
        Validate an integer string.

        :param x: the integer string
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().validate(x)
        if not all(x >= self.min_value):
            raise ValueError(
                f"All elements of x must be >= {self.min_value}, but "
                f"{x.min()} was encountered.")
        if not all(x <= self.max_value):
            raise ValueError(
                f"All elements of x must be <= {self.max_value}, but "
                f"{x.max()} was encountered.")

    def n_points(self) -> int:
        """
        Get the number of possible different integer strings.

        :return: (max_value - min_value + 1) ** dimension

        >>> print(IntSpace(4, -1, 3).n_points())
        625
        """
        return (self.max_value - self.min_value + 1) ** self.dimension

    def __str__(self) -> str:
        """
        Get the name of this integer space.

        :return: "ints" + dimension + dtype.char + min_value + "-" + max_value

        >>> print(IntSpace(4, -1, 3))
        ints4b-1-3
        """
        return f"ints{self.dimension}{self.dtype.char}" \
               f"{self.min_value}-{self.max_value}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_MIN, self.min_value)
        logger.key_value(KEY_MAX, self.max_value)



"""
The implementation of an Evolutionary Algorithm without crossover.

This is the basic `mu+lambda`-EA, but without using the binary crossover
operator. It works as follows:

1. Start with a population of `mu` random and `lambda` blank individuals.
2. In each iteration:
    2.1. Retain the best `mu` individuals (which will be the `mu` first
         individuals in the sorted population) and overwrite the `lambda`
         worse ones (which will be at indices `mu...mu+lambda-1`). Each of
         these individuals is overwritten with the results of the unary
         operator applied to one of the `mu` "parents". Each of the `mu`
         parents has the same chance to produce such a new "offspring", but no
         individual can be used as a parent again until all other `mu-1`
         selected individuals have produced at least one offspring. In other
         words, if `lambda > mu`, then each of the `mu` selected individuals
         will produce at least `lambda // mu` offspring and at most
         `1 + lambda // mu`.
    2.2. Shuffle the population to introduce randomness and fairness in the
         case that sorting is stable.
    2.3. Sort the population according to the objective value of the
         individuals. Ties are broken such that younger individuals are
         preferred over older ones (if they have the same objective value).

This EA only applies the unary search operator to sample new points in the
search space. Therefore, its population mainly guards against premature
convergence. If `mu=1` and `lambda=1`, then this algorithm is exactly
equivalent to the :class:`~moptipy.algorithms.rls.RLS` if the same unary and
nullary operator are used. It is only a bit slower due to the additional
overhead of representing a population as list of individual records.
"""
from typing import Final, Union, Callable, List, cast

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error
from moptipy.algorithms.utils import Individual, _no_shuffle, _no_random_int


# start book
class EAnoCR(Algorithm1):
    """
    The EA without crossover is a population-based metaheuristic.

    It starts with a population of `mu` random individuals. In each
    step, it retains the `mu` best solutions and generates `lambda`
    new offspring solutions from them using the unary search operator.
    From the joint set of `mu+lambda` solutions, it again selects the
    best `mu` ones for the next iteration. And so on.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the EA to an optimization problem.

        :param process: the black-box process object
        """
        mu: Final[int] = self.__mu  # mu: number of best solutions kept
        pop_size: Final[int] = mu + self.__lambda  # size = mu + lambda
        random: Final[Generator] = process.get_random()  # random gen
        # Omitted for brevity: store function references in variables
        # end book
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate
        ri: Final[Callable[[int], int]] = cast(    # only if m > 1, we
            Callable[[int], int], random.integers  # need random
            if mu > 1 else _no_random_int)         # indices
        shuffle: Final[Callable[[List], None]] = cast(
            Callable[[List], None], random.shuffle if mu > 1
            else _no_shuffle)  # shuffling only needed if mu > 1
        # start book
        # create population of mu random and lambda empty individuals
        pop: Final[List] = [None] * pop_size  # pre-allocate list
        f: Union[int, float] = 0  # variable to hold objective values
        for i in range(pop_size):  # fill population of size mu+lambda
            x = create()  # by creating point in search space
            if i < mu:  # only the first mu parents are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?
                    return   # computational budget exhausted -> quit
                f = evaluate(x)  # continue? ok, evaluate new solution
            pop[i] = Individual(x, f)  # create and store record

        gen: int = 1  # The first "real" generation has index 1
        while True:  # pop: keep 0..mu-1, overwrite mu..mu+lambda-1
            end: int = mu  # the end index for parents
            for oi in range(mu, pop_size):  # for all lambda offspring
                if should_terminate():  # only continue if we still...
                    return  # have sufficient budget ... otherwise quit
                offspring: Individual = pop[oi]  # pick offspring
                pi: int = ri(end)  # random parent in 0..end-1 = unused
                parent: Individual = pop[pi]  # pick parent record
                x = offspring.x  # the point in search space we work on
                op1(random, x, parent.x)  # apply unary operator to x
                offspring.f = evaluate(x)  # evaluate new point
                offspring.gen = gen  # mark as member of new generation

                end = end - 1  # decrease number of available parents
                if end != 0:  # swap parent to the end to not use again
                    pop[end], pop[pi] = parent, pop[end]  # parent->old
                else:  # oh: we have used all parents
                    end = mu  # then lambda >= mu and we re-use parents

            gen = gen + 1  # step generation counter

            shuffle(pop)  # ensure total fairness under stable sorting
            pop.sort()  # sort population: best individuals come first
# end book

    def __init__(self, op0: Op0, op1: Op1,
                 mu: int = 1, lambda_: int = 1) -> None:
        """
        Create the Evolutionary Algorithm (EA) without binary crossover.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        """
        super().__init__(f"eanocr_{mu}_{lambda_}", op0, op1)
        if not isinstance(mu, int):
            raise type_error(mu, "mu", int)
        if not (0 < mu <= 1_000_000):
            raise ValueError(f"invalid mu={mu}, must be in 1..1000000.")
        if not isinstance(lambda_, int):
            raise type_error(lambda_, "lambda", int)
        if not (0 < lambda_ <= 1_000_000):
            raise ValueError(
                f"invalid lambda={lambda_}, must be in 1..1000000.")
        #: the number of individuals to survive in each generation
        self.__mu: Final[int] = mu
        #: the number of offsprings per generation
        self.__lambda: Final[int] = lambda_

    def log_parameters_to(self, logger: KeyValueLogSection):
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("mu", self.__mu)
        logger.key_value("lambda", self.__lambda)



"""This module provides the class for implementing objective functions."""
from math import inf
from typing import Union, Callable, Optional, Final, Any

from moptipy.api import logging
from moptipy.api.component import CallableComponent, Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error


# start book
class Objective(Component):
    """
    An objective function subject to minimization.

    An objective function represents one optimization criterion that
    is used for rating the solution quality. All objective functions in
    our system are subject to minimization, meaning that smaller values
    are better.
    """

    def evaluate(self, x) -> Union[float, int]:
        """
        Evaluate a solution `x` and return its objective value.

        The return value is either an integer or a float and must be
        finite. Smaller objective values are better, i.e., all objective
        functions are subject to minimization.

        :param x: the candidate solution
        :return: the objective value
        """
    # end book

    def lower_bound(self) -> Union[float, int]:
        """
        Get the lower bound of the objective value.

        This function returns a theoretical limit for how good a solution
        could be at best. If no real limit is known, the function should
        return `-inf`.

        :return: the lower bound of the objective value
        """
        return -inf

    def upper_bound(self) -> Union[float, int]:
        """
        Get the upper bound of the objective value.

        This function returns a theoretical limit for how bad a solution could
        be at worst. If no real limit is known, the function should return
        `inf`.

        :return: the upper bound of the objective value
        """
        return inf


class CallableObjective(CallableComponent, Objective):
    """Wrapping a Callable such as a lambda into an objective function."""

    def __init__(self,
                 function: Callable[[Any], Union[float, int]],
                 lower_bound: Union[float, int] = -inf,
                 upper_bound: Union[float, int] = inf,
                 name: Optional[str] = None) -> None:
        """
        Create a wrapper mapping a Callable to an objective function.

        :param function: the function to wrap, can be a lambda expression
        :param lower_bound: the lower bound of the objective function
        :param upper_bound: the upper bound of the objective function
        :param name: the name of the objective function
        """
        super().__init__(inner=function,
                         name="unnamed_function" if (name is None) else name)

        if not (isinstance(lower_bound, (int, float))):
            raise type_error(lower_bound, "lower_bound", (int, float))

        if not (isinstance(upper_bound, (int, float))):
            raise type_error(upper_bound, "upper_bound", (int, float))

        if lower_bound >= upper_bound:
            raise ValueError(f"lower_bound {lower_bound} "
                             "must be less than upper_bound "
                             f"{upper_bound} but is not.")

        self.__lower_bound: Final[Union[int, float]] = lower_bound
        self.__upper_bound: Final[Union[int, float]] = upper_bound
        self.evaluate = self._inner  # type: ignore

    def lower_bound(self) -> Union[float, int]:
        """Get the lower bound of the objective value."""
        return self.__lower_bound

    def upper_bound(self) -> Union[float, int]:
        """Get the upper bound of the objective value."""
        return self.__upper_bound

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this function to the provided destination.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(logging.KEY_F_LOWER_BOUND, self.__lower_bound)
        logger.key_value(logging.KEY_F_UPPER_BOUND, self.__upper_bound)


def check_objective(objective: Objective) -> Objective:
    """
    Check whether an object is a valid instance of :class:`Objective`.

    :param objective: the objective
    :return: the object
    :raises TypeError: if `objective` is not an instance of
        :class:`Objective`
    """
    if not isinstance(objective, Objective):
        raise type_error(objective, "objective function", Objective)
    return objective



"""Utilities for interaction with numpy."""
from hashlib import sha512
from typing import Final, List, Iterable, cast
from typing import Set, Dict, Tuple

import numba  # type: ignore
import numpy
from numpy.random import default_rng, Generator, PCG64

from moptipy.utils.strings import str_to_bool
from moptipy.utils.types import type_error

#: A map associating all numpy integer types associated to tuples
#: of their respective minimum and maximum value.
__NP_INTS_MAP: Final[Dict[numpy.dtype, Tuple[int, int]]] = \
    {t: (int(ti.min), int(ti.max)) for t, ti in
     cast(List[Tuple[numpy.dtype, numpy.iinfo]],
          ((dtt, numpy.iinfo(dtt)) for dtt in
           cast(List[numpy.dtype], [numpy.dtype(dt) for dt in
                                    cast(List[type],
                                         [numpy.int8, numpy.uint8,
                                          numpy.int16, numpy.uint16,
                                          numpy.int32, numpy.uint32,
                                          numpy.int64, numpy.uint64])])))}

#: A tuple of integer types with their corresponding minimum and maximum
#: values in increasing orders of size. The tuple contains alternating
#: signed and unsigned types. It starts with the smallest signed type,
#: `numpy.int8` and ends with the largest unsigned type `numpy.uint64`.
#: If we have a range `[min..max]` of valids values, then we can look up this
#: range and find the integer type with the smallest memory footprint to
#: accommodate it. This is what :meth:`int_range_to_dtype` does.
__NP_INTS_LIST: Final[Tuple[Tuple[numpy.dtype, int, int], ...]] = \
    tuple([(a[3], a[1], a[2]) for a in
           sorted([(t.itemsize, a[0], a[1], t)
                   for t, a in __NP_INTS_MAP.items()])])

#: The default integer type: the signed 64 bit integer.
DEFAULT_INT: Final[numpy.dtype] = (__NP_INTS_LIST[-2])[0]

#: The default unsigned integer type: an unsigned 64 bit integer.
__DEFAULT_UNSIGNED_INT: Final[numpy.dtype] = (__NP_INTS_LIST[-1])[0]

#: The default boolean type.
DEFAULT_BOOL: Final[numpy.dtype] = numpy.dtype(numpy.bool_)

#: The default floating point type.
DEFAULT_FLOAT: Final[numpy.dtype] = numpy.zeros(1).dtype


def is_np_int(dtype: numpy.dtype) -> bool:
    """
    Check whether a :class:`numpy.dtype` is an integer type.

    :param dtype: the type

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import is_np_int
    >>> print(is_np_int(npx.dtype(npx.int8)))
    True
    >>> print(is_np_int(npx.dtype(npx.uint16)))
    True
    >>> print(is_np_int(npx.dtype(npx.float64)))
    False
    """
    return dtype.kind in ('i', 'u')


def is_np_float(dtype: numpy.dtype) -> bool:
    """
    Check whether a :class:`numpy.dtype` is a floating point type.

    :param dtype: the type

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import is_np_float
    >>> print(is_np_float(npx.dtype(npx.int8)))
    False
    >>> print(is_np_float(npx.dtype(npx.uint16)))
    False
    >>> print(is_np_float(npx.dtype(npx.float64)))
    True
    """
    return dtype.kind == 'f'


def int_range_to_dtype(min_value: int, max_value: int) -> numpy.dtype:
    """
    Convert an integer range to an appropriate numpy data type.

    The returned type is as compact as possible and signed types are
    preferred over unsigned types.
    The returned :class:`numpy.dtype` will allow accommodating all values in
    the inclusive interval `min_value..max_value`.

    :param min_value: the minimum value
    :param max_value: the maximum value
    :return: the numpy integer range
    :raises TypeError: if the parameters are not integers
    :raises ValueError: if the range is invalid

    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> print(int_range_to_dtype(0, 127))
    int8
    >>> print(int_range_to_dtype(0, 128))
    uint8
    >>> print(int_range_to_dtype(0, 32767))
    int16
    >>> print(int_range_to_dtype(0, 32768))
    uint16
    >>> print(int_range_to_dtype(0, (2 ** 31) - 1))
    int32
    >>> print(int_range_to_dtype(0, 2 ** 31))
    uint32
    >>> print(int_range_to_dtype(0, (2 ** 63) - 1))
    int64
    >>> print(int_range_to_dtype(0, 2 ** 63))
    uint64
    >>> print(int_range_to_dtype(0, (2 ** 64) - 1))
    uint64
    """
    if not isinstance(min_value, int):
        raise type_error(min_value, "min_value", int)
    if not isinstance(max_value, int):
        raise type_error(max_value, "max_value", int)
    if min_value > max_value:
        raise ValueError(
            f"min_value must be <= max_value, but min_value={min_value} "
            f"and max_value={max_value} was provided.")

    for t in __NP_INTS_LIST:
        if (min_value >= t[1]) and (max_value <= t[2]):
            return t[0]

    if min_value >= 0:
        raise ValueError(
            "max_value for unsigned integers must be <="
            f"{(__NP_INTS_LIST[-1])[2]}, but is {max_value} "
            f" for min_value={min_value}.")

    raise ValueError(
        f"Signed integer range cannot exceed {__NP_INTS_LIST[-2][1]}.."
        f"{__NP_INTS_LIST[-2][2]}, but {min_value}..{max_value} "
        "was specified.")


@numba.jit(forceobj=True)
def np_ints_max(shape, dtype: numpy.dtype = DEFAULT_INT) -> numpy.ndarray:
    """
    Create an integer array of the given length filled with the maximum value.

    :param shape: the requested shape
    :param dtype: the data type (defaults to 64bit integers)
    :return: the new array

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import np_ints_max
    >>> print(np_ints_max(4, npx.dtype("uint8")))
    [255 255 255 255]
    """
    return numpy.full(shape=shape,
                      fill_value=__NP_INTS_MAP[dtype][1],
                      dtype=dtype)


@numba.jit(forceobj=True)
def np_ints_min(shape, dtype: numpy.dtype = DEFAULT_INT) -> numpy.ndarray:
    """
    Create an integer array of the given length filled with the minimum value.

    :param shape: the requested shape
    :param dtype: the data type (defaults to 64 bit integers)
    :return: the new array

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import np_ints_min
    >>> print(np_ints_min(3, npx.dtype("int16")))
    [-32768 -32768 -32768]
    """
    return numpy.full(shape=shape,
                      fill_value=__NP_INTS_MAP[dtype][0],
                      dtype=dtype)


#: the default number of bytes for random seeds
__SEED_BYTES: Final = 8
#: the minimum acceptable random seed
__MIN_RAND_SEED: Final = 0
#: the maximum acceptable random seed
__MAX_RAND_SEED: Final = int((1 << (__SEED_BYTES * 8)) - 1)


def rand_seed_generate(random: Generator = default_rng()) -> int:
    """
    Generate a random seed.

    :param random: the random number generator to be used to generate the seed
    :return: the random seed
    :raises TypeError: if `random` is specified but is not an instance of
        `Generator`
    """
    if not isinstance(random, Generator):
        raise type_error(random, "random", Generator)
    return int.from_bytes(random.bytes(__SEED_BYTES),
                          byteorder='big', signed=False)


def rand_seed_check(rand_seed: int) -> int:
    """
    Make sure that a random seed is valid.

    :param rand_seed: the random seed to check
    :return: the rand seed

    :raises TypeError: if the random seed is not an `int`
    :raises ValueError: if the random seed is not valid
    """
    if not isinstance(rand_seed, int):
        raise type_error(rand_seed, "rand_seed", int)
    if (rand_seed < __MIN_RAND_SEED) or (rand_seed > __MAX_RAND_SEED):
        raise ValueError(f"rand_seed must be in {__MIN_RAND_SEED}.."
                         f"{__MAX_RAND_SEED}, but is {rand_seed}.")
    return rand_seed


def rand_generator(seed: int) -> Generator:
    """
    Instantiate a random number generator from a seed.

    :param seed: the random seed
    :return: the random number generator
    """
    return default_rng(rand_seed_check(seed))


def rand_seeds_from_str(string: str,
                        n_seeds: int) -> List[int]:
    """
    Reproducibly generate `n_seeds` unique random seeds from a `string`.

    This function will produce a sorted sequence of `n_seeds`random seeds,
    each of which being an unsigned 64 bit integer, from the string passed in.
    The same string will always yield the same sequence reproducibly.
    Running the function twice with different values of `n_seeds` will result
    in the two sets of random seeds, where the larger one (for the larger
    value of `n_seeds`) contains all elements of the smaller one.

    :param string: the string
    :param n_seeds: the number of seeds
    :return: a tuple of random seeds
    :raises TypeError: if the parameters do not follow the type contract
    :raises ValueError: if the parameter values are invalid

    >>> rand_seeds_from_str("hello world!", 1)
    [11688012229199056962]
    >>> rand_seeds_from_str("hello world!", 2)
    [3727742416375614079, 11688012229199056962]
    >>> rand_seeds_from_str("hello world!", 3)
    [3727742416375614079, 11688012229199056962, 17315292100125916507]

    >>> rand_seeds_from_str("metaheuristic optimization", 1)
    [12323230366215963648]
    >>> rand_seeds_from_str("metaheuristic optimization", 2)
    [12323230366215963648, 13673960948036381176]
    >>> rand_seeds_from_str("metaheuristic optimization", 3)
    [12323230366215963648, 13673960948036381176, 18426184104943646060]
    """
    if not isinstance(string, str):
        raise type_error(string, "string", str)
    if len(string) <= 0:
        raise ValueError("string must not be empty.")
    if not isinstance(n_seeds, int):
        raise type_error(n_seeds, "n_seeds", int)
    if n_seeds <= 0:
        raise ValueError(
            f"n_seeds must be positive, but is {n_seeds}.")

    seeds = bytearray(sha512(string.encode("utf8")).digest())
    seed1 = int.from_bytes(seeds[0:32], byteorder='big', signed=False)
    seed2 = int.from_bytes(seeds[32:64], byteorder='big', signed=False)
    del seeds

    # seed two PCG64 generators, each of which should use two 128 bit
    # numbers as seed
    g1 = Generator(PCG64(seed1))
    g2 = Generator(PCG64(seed2))

    generated: Set[int] = set()
    while len(generated) < n_seeds:
        g1, g2 = g2, g1
        generated.add(rand_seed_generate(g1))

    result = list(generated)
    result.sort()

    if len(result) != n_seeds:
        raise ValueError("Failed to generate {n_seeds} unique seeds.")
    return result


def strs_to_bools(lines: Iterable[str]) -> numpy.ndarray:
    """
    Convert an array of strings to a boolean numpy array.

    :param lines: the lines
    :return: the array

    >>> strs_to_bools(["T", "F", "T"])
    array([ True, False,  True])
    """
    return numpy.array([str_to_bool(s) for s in lines],
                       dtype=DEFAULT_BOOL)


def strs_to_uints(lines: Iterable[str]) -> numpy.ndarray:
    """
    Convert an array of strings to a numpy array of unsigned ints.

    :param lines: the lines
    :return: the array

    >>> strs_to_uints(["1", "2", "3"])
    array([1, 2, 3], dtype=uint64)
    """
    return numpy.array(lines, dtype=__DEFAULT_UNSIGNED_INT)


def strs_to_ints(lines: Iterable[str]) -> numpy.ndarray:
    """
    Convert an array of strings to a numpy array of signed ints.

    :param lines: the lines
    :return: the array

    >>> strs_to_ints(["-1", "2", "3"])
    array([-1,  2,  3])
    """
    return numpy.array(lines, dtype=DEFAULT_INT)


def strs_to_floats(lines: Iterable[str]) -> numpy.ndarray:
    """
    Convert an array of strings to a numpy array of floats.

    :param lines: the lines
    :return: the array

    >>> strs_to_floats(["-1.6", "2", "3"])
    array([-1.6,  2. ,  3. ])
    """
    return numpy.array(lines, dtype=DEFAULT_FLOAT)


@numba.njit(nogil=True)
def is_all_finite(a: numpy.ndarray) -> bool:
    """
    Check if an array is all finite.

    :param a: the input array
    :return: `True` if all elements in the array are finite, `False` otherwise

    >>> import numpy as npx
    >>> from moptipy.utils.nputils import is_all_finite
    >>> print(is_all_finite(npx.array([1.1, 2.1, 3])))
    True
    >>> print(is_all_finite(npx.array([1, 2, 3])))
    True
    >>> print(is_all_finite(npx.array([1.1, npx.inf, 3])))
    False
    """
    for x in a:
        if not numpy.isfinite(x):
            return False
    return True


#: the character identifying the numpy data type backing the space
KEY_NUMPY_TYPE: Final[str] = "dtype"



"""Functions that can be used to test objective functions."""
from math import isfinite, inf
from typing import Callable, Optional, Any, Union

from moptipy.api.objective import Objective, check_objective
from moptipy.api.space import Space
from moptipy.tests.component import validate_component


def validate_objective(
        objective: Objective,
        solution_space: Optional[Space] = None,
        make_solution_space_element_valid:
        Optional[Callable[[Any], Any]] = lambda x: x,
        is_deterministic: bool = True,
        lower_bound_threshold: Union[int, float] = -inf,
        upper_bound_threshold: Union[int, float] = inf,
        must_be_equal_to: Optional[
            Callable[[Any], Union[int, float]]] = None) -> None:
    """
    Check whether an object is a moptipy objective function.

    :param objective: the objective function to test
    :param solution_space: the solution space
    :param make_solution_space_element_valid: a function that makes an element
        from the solution space valid
    :param bool is_deterministic: is the objective function deterministic?
    :param lower_bound_threshold: the threshold for the lower bound
    :param upper_bound_threshold: the threshold for the upper bound
    :param must_be_equal_to: an optional function that should return the
        exactly same values as the objective function
    :raises ValueError: if `objective` is not a valid Objective
    """
    if not isinstance(objective, Objective):
        raise ValueError("Expected to receive an instance of Objective, but "
                         f"got a {type(objective)}.")
    check_objective(objective)
    validate_component(objective)

    if not (hasattr(objective, 'lower_bound')
            and callable(getattr(objective, 'lower_bound'))):
        raise ValueError("objective must have method lower_bound.")
    lower = objective.lower_bound()
    if not (isinstance(lower, (int, float))):
        raise ValueError("lower_bound() must return an int or float, but "
                         f"returned a {type(lower)}.")
    if (not isfinite(lower)) and (not (lower <= (-inf))):
        raise ValueError(
            f"lower bound must be finite or -inf, but is {lower}.")
    if lower < lower_bound_threshold:
        raise ValueError("lower bound must not be less than "
                         f"{lower_bound_threshold}, but is {lower}.")

    if not (hasattr(objective, 'upper_bound')
            and callable(getattr(objective, 'upper_bound'))):
        raise ValueError("objective must have method upper_bound.")
    upper = objective.upper_bound()
    if not (isinstance(upper, (int, float))):
        raise ValueError("upper_bound() must return an int or float, but "
                         f"returned a {type(upper)}.")
    if (not isfinite(upper)) and (not (upper >= inf)):
        raise ValueError(
            f"upper bound must be finite or +inf, but is {upper}.")
    if upper > upper_bound_threshold:
        raise ValueError(
            f"upper bound must not be more than {upper_bound_threshold}, "
            f"but is {lower}.")

    if lower >= upper:
        raise ValueError("Result of lower_bound() must be smaller than "
                         f"upper_bound(), but got {lower} vs. {upper}.")

    count: int = 0
    if make_solution_space_element_valid is not None:
        count += 1
    if solution_space is not None:
        count += 1
    if count <= 0:
        return
    if count < 2:
        raise ValueError("either provide both of solution_space and "
                         "make_solution_space_element_valid or none.")

    x = solution_space.create()
    if x is None:
        raise ValueError("solution_space.create() produced None.")
    x = make_solution_space_element_valid(x)
    if x is None:
        raise ValueError("make_solution_space_element_valid() produced None.")
    solution_space.validate(x)

    if not (hasattr(objective, 'evaluate')
            and callable(getattr(objective, 'evaluate'))):
        raise ValueError("objective must have method evaluate.")
    res = objective.evaluate(x)
    if not (isinstance(res, (int, float))):
        raise ValueError(f"evaluate(x) of {x} must return an int or float, "
                         f"but returned a {type(res)}.")

    if (res < lower) or (res > upper):
        raise ValueError(f"evaluate(x) of {x} must return a value in"
                         "[lower_bound(), upper_bound()], but returned "
                         f"{res} vs. [{lower},{upper}].")
    if must_be_equal_to is not None:
        exp = must_be_equal_to(x)
        if exp != res:
            raise ValueError(f"expected to get {exp}, but got {res}.")

    res2 = objective.evaluate(x)
    if not (isinstance(res2, (int, float))):
        raise ValueError(f"evaluate(x) of {x} must return an int or float, "
                         f"but returned a {type(res2)}.")

    if (res2 < lower) or (res2 > upper):
        raise ValueError(f"evaluate(x) of {x} must return a value in"
                         "[lower_bound(), upper_bound()], but returned "
                         f"{res2} vs. [{lower},{upper}].")

    if is_deterministic and (res != res2):
        raise ValueError(f"evaluating {x} twice yielded the two different "
                         f"results {res} and {res2}!")



"""Provides the base class for all components of the moptipy API."""
from typing import Callable, Final

from moptipy.api import logging
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_name_of, type_error


class Component:
    """The base class for all components of the moptipy API."""

    def __repr__(self):
        """
        Get the string representation of this object.

        :return: the value returned by :meth:`__str__`
        """
        return str(self)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        logger.key_value(logging.KEY_NAME, str(self))
        logger.key_value(logging.KEY_CLASS, type_name_of(self))


class CallableComponent(Component):
    """Wrap a Callable such as a lambda into a component."""

    def __init__(self,
                 inner: Callable,
                 name: str) -> None:
        """
        Create a wrapper mapping a Callable to an component.

        :param inner: the function to wrap, e.g., a lambda
        :param name: the name of the component
        :raises TypeError: if `inner` is not callable
        """
        if not callable(inner):
            raise type_error(inner, "inner function", call=True)

        #: the inner callable
        self._inner: Final[Callable] = inner
        #: the name of the component
        self.__name: Final[str] = sanitize_name(name)

    def __str__(self) -> str:
        """
        Get the name of this component.

        :return: the name
        """
        return self.__name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(logging.KEY_INNER_CLASS, type_name_of(self._inner))



"""
The implementation of the hill climbing algorithm with restarts `hcr`.

This algorithm basically works like the normal hill climber `hc`
(:class:`~moptipy.algorithms.hill_climber.HillClimber`), but it will restart
automatically if no move was successful for
`max_moves_without_improvement` iterative steps. It therefore maintains an
internal counter `count` which is set to zero at the beginning of each restart
and which is also set to zero again any time a move successfully improved the
best-so-far solution of the current restart. If a search move, i.e., an
application of the unary operator, yielded a new solution which is not better
than the best-so-far solution of the current restart, `count` is incremented.
If `count >= max_moves_without_improvement`, the algorithm begins a new
restart with a new random solution.
"""
from typing import Final, Union, Callable

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.types import type_error
from moptipy.utils.logger import KeyValueLogSection


# start book
class HillClimberWithRestarts(Algorithm1):
    """
    The stochastic hill climbing algorithm only accepts improving moves.

    In each step, a hill climber creates a modified copy `new_x` of the
    current best solution `best_x`. If `new_x` is better than `best_x`,
    it becomes the new `best_x`. Otherwise, it is discarded. If no
    improvement is made for `max_moves_without_improvement` steps, the
    algorithm restarts.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the hill climber with restarts to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate
        limit: Final[int] = self.max_moves_without_improvement

        while not should_terminate():  # Until we need to quit....
            self.op0.op0(random, best_x)  # Create random solution and
            best_f: Union[int, float] = evaluate(best_x)
            count: int = 0  # The counter of unsuccessful moves = 0.

            while not should_terminate():  # Until we need to quit...
                op1(random, new_x, best_x)  # new_x=neighbor of best_x
                new_f: Union[int, float] = evaluate(new_x)
                if new_f < best_f:  # new_x is _better_ than best_x?
                    best_f = new_f  # Store its objective value.
                    best_x, new_x = new_x, best_x  # Swap best and new.
                    count = 0  # Reset unsuccessful move counter.
                else:  # The move did not lead to an improvement!?
                    count += 1  # Increment unsuccessful move counter.
                    if count >= limit:  # Too many unsuccessful moves?
                        break  # Break inner loop, start again randomly.
# end book

    def __init__(self, op0: Op0, op1: Op1,
                 max_moves_without_improvement: int) -> None:
        """
        Create the hill climber.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param max_moves_without_improvement: the maximum number of
            moves without improvement before a restart
        """
        super().__init__(
            f"hcr_{max_moves_without_improvement}", op0, op1)
        if not isinstance(max_moves_without_improvement, int):
            raise type_error(max_moves_without_improvement,
                             "max_moves_without_improvement", int)
        if max_moves_without_improvement <= 0:
            raise ValueError(
                "max_moves_without_improvement must be positive, but"
                f" is {max_moves_without_improvement}.")
        #: the maximum moves without improvement
        self.max_moves_without_improvement: Final[int] = \
            max_moves_without_improvement

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("maxMovesWithoutImprovement",
                         self.max_moves_without_improvement)



"""The markdown text format driver."""

from io import TextIOBase
from typing import Final

from moptipy.utils.text_format import TextFormatDriver


class Markdown(TextFormatDriver):
    r"""
    The markdown text driver.

    >>> from io import StringIO
    >>> from moptipy.utils.table import Table
    >>> from moptipy.utils.text_format import FormattedStr
    >>> s = StringIO()
    >>> md = Markdown.instance()
    >>> print(str(md))
    md
    >>> with Table(s, "lrc", md) as t:
    ...     with t.header() as h:
    ...         h.cell(FormattedStr("1", bold=True))
    ...         h.cell(FormattedStr("2", code=True))
    ...         h.cell(FormattedStr("3", italic=True))
    ...     with t.section() as g:
    ...         with g.row() as r:
    ...             r.cell("a")
    ...             r.cell("b")
    ...             r.cell("c")
    ...         with g.row() as r:
    ...             r.cell("d")
    ...             r.cell("e")
    ...             r.cell("f")
    >>> print(f"'{s.getvalue()}'")
    '|**1**|`2`|*3*|
    |:--|--:|:-:|
    |a|b|c|
    |d|e|f|
    '
    """

    def end_table_header(self, stream: TextIOBase, cols: str) -> None:
        """End the header of a markdown table."""
        for c in cols:
            if c == "l":
                stream.write("|:--")
            elif c == "c":
                stream.write("|:-:")
            elif c == "r":
                stream.write("|--:")
            else:
                raise ValueError(f"Invalid col '{c}' in '{cols}'.")
        stream.write("|\n")

    def end_table_row(self, stream: TextIOBase, cols: str,
                      section_index: int, row_index: int) -> None:
        """End a row in a Markdown table."""
        stream.write("|\n")

    def begin_table_cell(self, stream: TextIOBase, cols: str,
                         section_index: int, row_index: int,
                         col_index: int) -> None:
        """Begin a Markdown table cell."""
        stream.write("|")

    def text(self, stream: TextIOBase, text: str, bold: bool, italic: bool,
             code: bool) -> None:
        """Print a text string."""
        if len(text) <= 0:
            return
        if bold:
            stream.write("**")
        if italic:
            stream.write("*")
        if code:
            stream.write("`")
        stream.write(text)
        if code:
            stream.write("`")
        if italic:
            stream.write("*")
        if bold:
            stream.write("**")

    def __str__(self):
        """
        Get the appropriate file suffix.

        :returns: the file suffix
        :retval 'md': always
        """
        return "md"

    def render_numeric_exponent(self, e: str) -> str:
        """
        Render the numerical exponent in markdown.

        :param e: the exponent
        :returns: the rendered exponent
        :retval: `*10^{e}^`
        """
        return f"*10^{e}^"

    @staticmethod
    def instance() -> 'Markdown':
        """
        Get the markdown format singleton instance.

        :returns: the singleton instance of the Markdown format
        """
        attr: Final[str] = "_instance"
        func: Final = Markdown.instance
        if not hasattr(func, attr):
            setattr(func, attr, Markdown())
        return getattr(func, attr)



"""
Approximate the expected running time to reach certain goals.

The (empirically estimated) Expected Running Time (ERT) tries to give an
impression of how long an algorithm needs to reach a certain solution quality.
"""

from dataclasses import dataclass
from math import isfinite, inf
from typing import Optional, Iterable, List, Final, cast, Union, Any, Dict, \
    Callable

import numpy as np

import moptipy.api.logging as lg
import moptipy.utils.nputils as npu
from moptipy.evaluation._utils import _get_reach_index
from moptipy.evaluation.base import MultiRun2DData, F_NAME_SCALED, \
    F_NAME_NORMALIZED, F_NAME_RAW, KEY_N
from moptipy.evaluation.progress import Progress
from moptipy.utils.console import logger
from moptipy.utils.logger import CSV_SEPARATOR, KEY_VALUE_SEPARATOR, \
    COMMENT_CHAR
from moptipy.utils.path import Path
from moptipy.utils.strings import num_to_str
from moptipy.utils.types import type_error


def compute_single_ert(source: Iterable[Progress],
                       goal_f: Union[int, float]) -> float:
    """
    Compute a single ERT.

    The ERT is the sum of the time that the runs spend with a
    best-so-far quality greater or equal than `goal_f` divided by the
    number of runs that reached `goal_f`. The idea is that the
    unsuccessful runs spent their complete computational budget and
    once they have terminated, we would immediately start a new,
    independent run.

    :param source: the source array
    :param goal_f: the goal objective value
    :return: the ERT
    """
    n_success: int = 0
    time_sum: int = 0
    for progress in source:
        size = progress.f.size
        idx = size - _get_reach_index(progress.f, goal_f)
        if idx < size:
            n_success += 1
        else:
            idx = cast(np.integer, -1)
        time_sum += progress.time[idx]
    if time_sum <= 0:
        raise ValueError(f"Time sum cannot be {time_sum}.")
    if n_success <= 0:
        return inf
    return time_sum / n_success


@dataclass(frozen=True, init=False, order=True)
class Ert(MultiRun2DData):
    """Estimate the Expected Running Time (ERT)."""

    #: The ert function
    ert: np.ndarray

    def __init__(self,
                 algorithm: Optional[str],
                 instance: Optional[str],
                 n: int,
                 time_unit: str,
                 f_name: str,
                 ert: np.ndarray):
        """
        Create the Ert function.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param n: the total number of runs
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        :param ert: the ert matrix
        """
        super().__init__(algorithm, instance, n, time_unit, f_name)
        if not isinstance(ert, np.ndarray):
            raise type_error(ert, "ert", np.ndarray)
        ert.flags.writeable = False

        f: Final[np.ndarray] = ert[:, 0]
        if not npu.is_all_finite(f):
            raise ValueError(
                f"Ert x-axis must be all finite, but encountered {f}.")
        ll = f.size
        if ll > 1:
            if np.any(f[1:] <= f[:-1]):
                raise ValueError("f data must be strictly increasing,"
                                 f"but encountered {f}.")

        t: Final[np.ndarray] = ert[:, 1]
        if np.isfinite(t[0]) or (np.isposinf(t[0])):
            if not npu.is_all_finite(t[1:]):
                raise ValueError(
                    "non-first ert y-axis elements must be all finite.")
            if np.any(t[1:] >= t[:-1]):
                raise ValueError("t data must be strictly decreasing,"
                                 f"but encountered {t}.")
        else:
            raise ValueError(
                f"first ert y-axis element cannot be {t[0]}.")
        object.__setattr__(self, "ert", ert)

    def to_csv(self, file: str,
               put_header: bool = True) -> Path:
        """
        Store a :class:`Ert` record in a CSV file.

        :param file: the file to generate
        :param put_header: should we put a header with meta-data?
        :return: the fully resolved file name
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing ERT to CSV file '{path}'.")

        with path.open_for_write() as out:
            sep: Final[str] = CSV_SEPARATOR
            if put_header:
                kv: Final[str] = KEY_VALUE_SEPARATOR
                cmt: Final[str] = COMMENT_CHAR
                if self.algorithm is not None:
                    out.write(
                        f"{cmt} {lg.KEY_ALGORITHM}{kv}{self.algorithm}\n")
                if self.instance is not None:
                    out.write(
                        f"{cmt} {lg.KEY_INSTANCE}{kv}{self.instance}\n")
                out.write(
                    f"{cmt} {KEY_N}{kv}{self.n}\n")
            out.write(f"{self.f_name}{sep}ert[{self.time_unit}]\n")
            for v in self.ert:
                out.write(
                    f"{num_to_str(v[0])}{sep}{num_to_str(v[1])}\n")

        logger(f"Done writing ERT to CSV file '{path}'.")

        path.enforce_file()
        return path

    @staticmethod
    def create(source: Iterable[Progress],
               f_lower_bound: Union[int, float, Callable, None] = None,
               use_default_lower_bounds: bool = True) -> 'Ert':
        """
        Create one single Ert record from an iterable of Progress records.

        :param source: the set of progress instances
        :param f_lower_bound: the lower bound for the objective value, or a
            callable that is applied to a progress object to get the lower
            bound
        :param use_default_lower_bounds: should we use the default lower
            bounds
        :return: the Ert record
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)

        lower_bound: Union[int, float] = inf
        if f_lower_bound is not None:
            if not callable(f_lower_bound):
                if not isfinite(f_lower_bound):
                    raise ValueError("f_lower_bound must be finite "
                                     f"but is {f_lower_bound}.")
                lower_bound = f_lower_bound
                f_lower_bound = None

        algorithm: Optional[str] = None
        instance: Optional[str] = None
        time_unit: Optional[str] = None
        f_name: Optional[str] = None
        f_list: List[np.ndarray] = []
        n: int = 0

        prgs: Final[List[Progress]] = cast(List[Progress], source) \
            if isinstance(source, list) else list(source)

        for progress in prgs:
            if not isinstance(progress, Progress):
                raise type_error(progress, "progress input", Progress)
            if n <= 0:
                algorithm = progress.algorithm
                instance = progress.instance
                time_unit = progress.time_unit
                f_name = progress.f_name
            else:
                if algorithm != progress.algorithm:
                    algorithm = None
                if instance != progress.instance:
                    instance = None
                if time_unit != progress.time_unit:
                    raise ValueError(
                        f"Cannot mix time units {time_unit} "
                        f"and {progress.time_unit}.")
                if f_name != progress.f_name:
                    raise ValueError(f"Cannot mix f-names {f_name} "
                                     f"and {progress.f_name}.")
            n += 1
            if use_default_lower_bounds and \
                    (progress.f_standard is not None) and \
                    (progress.f_name == F_NAME_RAW):
                if lower_bound > progress.f_standard:
                    lower_bound = progress.f_standard
            if f_lower_bound is not None:
                lb = f_lower_bound(progress)
                if not isinstance(lb, (int, float)):
                    raise type_error(lb, "computed lower bound", (int, float))
                if not isfinite(lb):
                    raise ValueError(f"Invalid computed lower bound {lb}.")
                if lb < lower_bound:
                    lower_bound = lb
            f_list.append(progress.f)

        if n <= 0:
            raise ValueError("Did not encounter any progress information.")

        if use_default_lower_bounds:
            if f_name == F_NAME_SCALED:
                lower_bound = min(lower_bound, 1)
            elif f_name == F_NAME_NORMALIZED:
                lower_bound = min(lower_bound, 0)

        # get unique x-values and make sure that lower bound is included
        has_lb: Final[bool] = isfinite(lower_bound)
        if has_lb:
            f_list.append(np.array([lower_bound]))
        x = np.concatenate(f_list)
        del f_list
        x = np.unique(x)
        if has_lb:
            x = x[x >= lower_bound]
        base_len: Final[int] = x.size

        # prepare for backward iteration over arrays
        indices = np.array([(ppr.f.size - 1) for ppr in prgs],
                           dtype=npu.DEFAULT_INT)
        y = np.empty(base_len, dtype=npu.DEFAULT_FLOAT)

        for out_index in range(base_len):
            f_lim = x[out_index]
            found = 0
            time_sum: int = 0
            for r_idx, pr in enumerate(prgs):
                idx = indices[r_idx]
                f_vals = pr.f
                # did we fulfill the limit?
                if f_vals[idx] <= f_lim:
                    found += 1
                    # can we move time cursor back?
                    if (idx > 0) and (f_vals[idx - 1] <= f_lim):
                        idx -= 1
                        indices[r_idx] = idx
                else:
                    # condition not fulfilled, need to use maximum time
                    idx = -1
                time_sum += pr.time[idx]

            # compute ert value: infinite if no run meets condition
            y[out_index] = inf if (found <= 0) else time_sum / found

        # convert the two arrays into one matrix
        ert = np.concatenate((x, y)).reshape((base_len, 2), order='F')

        return Ert(algorithm, instance, n,
                   time_unit, f_name, ert)

    @staticmethod
    def from_progresses(source: Iterable[Progress],
                        consumer: Callable[['Ert'], Any],
                        f_lower_bound: Optional[float] = None,
                        use_default_lower_bounds: bool = True,
                        join_all_algorithms: bool = False,
                        join_all_instances: bool = False) -> None:
        """
        Compute one or multiple ERTs from a stream of end results.

        :param source: the set of progress instances
        :param f_lower_bound: the lower bound for the objective value
        :param use_default_lower_bounds: should we use the default lower
            bounds
        :param consumer: the destination to which the new records will be
            passed
        :param join_all_algorithms: should the Ert be aggregated over all
            algorithms
        :param join_all_instances: should the Ert be aggregated over all
            algorithms
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(join_all_algorithms, bool):
            raise type_error(join_all_algorithms, "join_all_algorithms", bool)
        if not isinstance(join_all_instances, bool):
            raise type_error(join_all_instances, "join_all_instances", bool)

        if join_all_algorithms and join_all_instances:
            consumer(Ert.create(source, f_lower_bound,
                                use_default_lower_bounds))
            return

        sorter: Dict[str, List[Progress]] = {}
        for er in source:
            if not isinstance(er, Progress):
                raise type_error(er, "progress source", Progress)
            key = er.instance if join_all_algorithms else \
                er.algorithm if join_all_instances else \
                f"{er.algorithm}/{er.instance}"
            if key in sorter:
                lst = sorter[key]
            else:
                lst = []
                sorter[key] = lst
            lst.append(er)

        if len(sorter) <= 0:
            raise ValueError("source must not be empty")

        if len(sorter) > 1:
            keyz = list(sorter.keys())
            keyz.sort()
            for key in keyz:
                consumer(Ert.create(sorter[key], f_lower_bound,
                                    use_default_lower_bounds))
        else:
            consumer(Ert.create(next(iter(sorter.values())), f_lower_bound,
                                use_default_lower_bounds))



"""Some internal utilities for optimization."""
from typing import Final, Union, List


# start individual
class Individual:
    """
    A point in the search space, its quality, and creation time.

    An Individual record stores a point in the search space :attr:`x`
    together with the corresponding objective value :attr:`f`. It also
    stores a "generation index" :attr:`gen`, i.e., the time when the
    point was created or modified.

    This allows for representing and storing solutions in a population.
    If the population is sorted, then individual records with better
    objective value will be moved to the beginning of the list. Ties are
    broken such that younger individuals (with higher :attr:`gen` value)
    are preferred.
    """

    def __init__(self, x, f: Union[int, float]):
        """
        Create the individual record.

        :param x: the data structure for a point in the search space
        :param f: the corresponding objective value
        """
        #: the point in the search space
        self.x: Final = x
        #: the objective value corresponding to x
        self.f: Union[int, float] = f
        #: the generation index when the record was created
        self.gen: int = 0

    def __lt__(self, other):
        """Precedence if 1) better or b) equally good but younger."""
        f1: Final[Union[int, float]] = self.f
        f2: Final[Union[int, float]] = other.f
        return (f1 < f2) or ((f1 == f2) and (self.gen > other.gen))
# end individual

    def __str__(self) -> str:
        """
        Return the string representation of this object.

        :returns: the string representation of this object.
        """
        return f"({self.f}, {self.gen})"


def _no_random_int(_: int) -> int:
    """Do nothing."""
    return 0


def _no_shuffle(_: List) -> None:
    """Do nothing."""



"""Test stuff on bit strings."""

from typing import Callable, Union, Iterable, List, Optional, Dict, Any, \
    cast, Final

from numpy.random import default_rng

from moptipy.api.algorithm import Algorithm
from moptipy.api.objective import Objective
from moptipy.api.operators import Op0, Op1
from moptipy.examples.bitstrings.leadingones import LeadingOnes
from moptipy.examples.bitstrings.onemax import OneMax
from moptipy.spaces.bitstrings import BitStrings
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.op0 import validate_op0
from moptipy.tests.op1 import validate_op1
from moptipy.utils.types import type_error


def dimensions_for_tests() -> Iterable[int]:
    """
    Get a sequence of dimensions for tests.

    :returns: the sequence of integers
    """
    r = default_rng()
    bs: List[int] = [1, 2, 3, 4, 5, 10, 16, 100,
                     int(r.integers(20, 50)), int(r.integers(200, 300))]
    r.shuffle(cast(List, bs))
    return bs


def bitstrings_for_tests() -> Iterable[BitStrings]:
    """
    Get a sequence of bit strings for tests.

    :returns: the sequence of BitStrings
    """
    return [BitStrings(i) for i in dimensions_for_tests()]


def validate_op0_on_1_bitstrings(
        op0: Union[Op0, Callable[[BitStrings], Op0]],
        search_space: BitStrings,
        number_of_samples: Optional[int] = None,
        min_unique_samples:
        Optional[Union[int, Callable[[int, BitStrings], int]]]
        = None) -> None:
    """
    Validate the unary operator on one bit strings instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: Dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_bitstrings(
        op0: Union[Op0, Callable[[BitStrings], Op0]],
        number_of_samples: Optional[int] = None,
        min_unique_samples:
        Optional[Union[int, Callable[[int, BitStrings], int]]]
        = None) -> None:
    """
    Validate the unary operator on several BitStrings instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for bst in bitstrings_for_tests():
        validate_op0_on_1_bitstrings(op0, bst,
                                     number_of_samples, min_unique_samples)


def validate_op1_on_1_bitstrings(
        op1: Union[Op1, Callable[[BitStrings], Op1]],
        search_space: BitStrings,
        number_of_samples: Optional[int] = None,
        min_unique_samples:
        Optional[Union[int, Callable[[int, BitStrings], int]]]
        = None) -> None:
    """
    Validate the unary operator on one BitStrings instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: Dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1(**args)


def validate_op1_on_bitstrings(
        op1: Union[Op1, Callable[[BitStrings], Op1]],
        number_of_samples: Optional[int] = None,
        min_unique_samples:
        Optional[Union[int, Callable[[int, BitStrings], int]]]
        = None) -> None:
    """
    Validate the unary operator on several BitStrings instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for bst in bitstrings_for_tests():
        validate_op1_on_1_bitstrings(op1, bst,
                                     number_of_samples, min_unique_samples)


def validate_algorithm_on_bitstrings(
        objective: Union[Objective, Callable[[int], Objective]],
        algorithm: Union[Algorithm,
                         Callable[[BitStrings], Algorithm]],
        dimension: int = 5,
        max_fes: int = 100,
        required_result: Optional[Union[int, Callable[
            [int, int], int]]] = None) -> None:
    """
    Check the validity of a black-box algorithm on a bit strings problem.

    :param algorithm: the algorithm or algorithm factory
    :param objective: the instance or instance factory
    :param dimension: the dimension of the problem
    :param max_fes: the maximum number of FEs
    :param required_result: the optional required result quality
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, 'algorithm', Algorithm, True)
    if not (isinstance(objective, Objective) or callable(objective)):
        raise type_error(objective, "objective", Objective, True)
    if not isinstance(dimension, int):
        raise type_error(dimension, 'dimension', int)
    if dimension <= 0:
        raise ValueError(f"dimension must be > 0, but got {dimension}.")

    if callable(objective):
        objective = objective(dimension)
    if not isinstance(objective, Objective):
        raise type_error(objective, "result of callable 'objective'",
                         Objective)
    bs: Final[BitStrings] = BitStrings(dimension)
    if callable(algorithm):
        algorithm = algorithm(bs)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "result of callable 'algorithm'",
                         Algorithm)

    goal: Optional[int]
    if callable(required_result):
        goal = required_result(max_fes, dimension)
    else:
        goal = required_result

    validate_algorithm(algorithm=algorithm,
                       solution_space=bs,
                       objective=objective,
                       max_fes=max_fes,
                       required_result=goal)


def validate_algorithm_on_onemax(
        algorithm: Union[Algorithm,
                         Callable[[BitStrings], Algorithm]]) -> None:
    """
    Check the validity of a black-box algorithm on onemax.

    :param algorithm: the algorithm or algorithm factory
    """
    max_fes: Final[int] = 100
    for i in dimensions_for_tests():
        rr: int
        if i < 3:
            rr = 0
        else:
            rr = max(i // 2, i - int(max_fes ** 0.5))
        validate_algorithm_on_bitstrings(
            objective=OneMax,
            algorithm=algorithm,
            dimension=i,
            max_fes=max_fes,
            required_result=rr)


def validate_algorithm_on_leadingones(
        algorithm: Union[Algorithm,
                         Callable[[BitStrings], Algorithm]]) -> None:
    """
    Check the validity of a black-box algorithm on leadingones.

    :param algorithm: the algorithm or algorithm factory
    """
    max_fes: Final[int] = 100
    for i in dimensions_for_tests():
        rr: int
        if i <= 3:
            rr = 0
        elif max_fes > (10 * (i ** 1.1)):
            rr = i - 2
        elif max_fes > (4 * (i ** 1.15)):
            rr = i - 1
        else:
            rr = i
        validate_algorithm_on_bitstrings(
            objective=LeadingOnes,
            algorithm=algorithm,
            dimension=i,
            max_fes=max_fes,
            required_result=rr)



"""A process with logging, where search and solution space are the same."""
from typing import Optional, Union, List, Final, cast

from moptipy.api import logging
from moptipy.api._process_base import _ns_to_ms, _TIME_IN_NS
from moptipy.api._process_no_ss import _ProcessNoSS
from moptipy.api.algorithm import Algorithm
from moptipy.api.objective import Objective
from moptipy.api.space import Space
from moptipy.utils.logger import Logger
from moptipy.utils.path import Path
from moptipy.utils.types import type_error


class _ProcessNoSSLog(_ProcessNoSS):
    """An process with logging, with the same search and solution space."""

    def __init__(self,
                 solution_space: Space,
                 objective: Objective,
                 algorithm: Algorithm,
                 log_file: Path,
                 rand_seed: Optional[int] = None,
                 max_fes: Optional[int] = None,
                 max_time_millis: Optional[int] = None,
                 goal_f: Union[int, float, None] = None,
                 log_all_fes: bool = False) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the search- and solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the process
            is terminated
        :param log_all_fes: should we log all FEs?
        """
        super().__init__(solution_space=solution_space,
                         objective=objective,
                         algorithm=algorithm,
                         log_file=log_file,
                         rand_seed=rand_seed,
                         max_fes=max_fes,
                         max_time_millis=max_time_millis,
                         goal_f=goal_f)
        if not isinstance(log_file, str):
            raise type_error(log_file, "log_file", str)
        if not isinstance(log_all_fes, bool):
            raise type_error(log_all_fes, "log_all_fes", bool)

        #: `True` if all FEs are logged, `False` to only log improvements.
        self.__log_all: Final[bool] = log_all_fes
        #: The in-memory log
        self.__log: List[List[Union[int, float]]] = []
        #: the quick access to the log appending method
        self.__log_append = self.__log.append

    def evaluate(self, x) -> Union[float, int]:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError('The process has been terminated and '
                                 'the algorithm knows it.')
            return self._current_best_f

        result: Final[Union[int, float]] = self._f(x)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes
        do_log: bool = self.__log_all
        ctn: int = 0

        if result < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = result
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (result <= self._end_f)
            self._copy_y(self._current_best_y, x)
            do_log = True

        if do_log:
            if ctn <= 0:
                self._current_time_nanos = ctn = _TIME_IN_NS()
            self.__log_append([current_fes, ctn, result])

        if do_term:
            self.terminate()

        return result

    def register(self, x, f: Union[int, float]) -> None:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError('The process has been terminated and '
                                 'the algorithm knows it.')
            return

        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes
        do_log: bool = self.__log_all
        ctn: int = 0

        if f < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = f
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (f <= self._end_f)
            self._copy_y(self._current_best_y, x)
            do_log = True

        if do_log:
            if ctn <= 0:
                self._current_time_nanos = ctn = _TIME_IN_NS()
            self.__log_append([current_fes, ctn, f])

        if do_term:
            self.terminate()

    def _check_timing(self) -> None:
        super()._check_timing()
        last_time: int = -1
        last_fe: int = -1
        start_time: Final[int] = self._start_time_nanos
        current_time: Final[int] = self._current_time_nanos
        for row in self.__log:
            fes: int = cast(int, row[0])
            time: int = cast(int, row[1])
            if fes < last_fe:
                raise ValueError(f"fe={fes} after fe={last_fe}?")
            if time < last_time:
                raise ValueError(
                    f"time={time} of fe={fes} is less than "
                    f"last_time={last_time} of last_fe={last_fe}")
            if time < start_time:
                raise ValueError(
                    f"time={time} of fe={fes} is less than "
                    f"start_time_nanos={start_time}")
            if time > current_time:
                raise ValueError(
                    f"time={time} of fe={fes} is greater than "
                    f"current_time_nanos={current_time}")
            last_time = time
            last_fe = fes

    def _write_log(self, logger: Logger) -> None:
        if len(self.__log) > 0:
            with logger.csv(logging.SECTION_PROGRESS,
                            [logging.PROGRESS_FES,
                             logging.PROGRESS_TIME_MILLIS,
                             logging.PROGRESS_CURRENT_F]) as csv:
                for row in self.__log:
                    csv.row([row[0], _ns_to_ms(cast(int, row[1])
                                               - self._start_time_nanos),
                             row[2]])
        del self.__log
        super()._write_log(logger)

    def __str__(self) -> str:
        return "LoggingProcessWithoutSearchSpace"



"""An internal module with the base class for implementing Processes."""
from math import inf
from threading import Lock, Timer
from time import time_ns
from typing import Optional, Final, Union, Callable

from moptipy.api.logging import _ALL_SECTIONS, KEY_MAX_FES, \
    KEY_MAX_TIME_MILLIS, SCOPE_PROCESS, KEY_GOAL_F
from moptipy.api.process import Process, check_goal_f, check_max_fes, \
    check_max_time_millis
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.logger import SECTION_START, SECTION_END
from moptipy.utils.types import type_error

# the function used to get the time
_TIME_IN_NS: Final[Callable[[], int]] = time_ns


def _ns_to_ms(nanos: int) -> int:
    """
    Convert nanoseconds to milliseconds by rounding up.

    :param int nanos: the nanoseconds
    :returns: the corresponding milliseconds, rounded up

    >>> _ns_to_ms(0)
    0
    >>> _ns_to_ms(1)
    1
    >>> _ns_to_ms(999_999)
    1
    >>> _ns_to_ms(1_000_000)
    1
    >>> _ns_to_ms(1_000_001)
    2
    >>> _ns_to_ms(1_999_999)
    2
    >>> _ns_to_ms(2_000_000)
    2
    >>> _ns_to_ms(2_000_001)
    3
    """
    return (nanos + 999_999) // 1_000_000


class _ProcessBase(Process):
    """The internal base class for implementing optimization processes."""

    def __init__(self,
                 max_fes: Optional[int] = None,
                 max_time_millis: Optional[int] = None,
                 goal_f: Union[int, float, None] = None) -> None:
        """
        Initialize information that every black-box process must have.

        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the process
            is terminated
        """
        super().__init__()
        #: This will be `True` after :meth:`terminate` has been called.
        self._terminated: bool = False  # +book
        #: This becomes `True` when :meth:`should_terminate` returned `True`.
        self._knows_that_terminated: bool = False
        #: The time when the process was started, in nanoseconds.
        self._start_time_nanos: Final[int] = _TIME_IN_NS()
        #: The internal lock, needed to protect :meth:`terminate`.
        self.__lock: Final[Lock] = Lock()
        #: The maximum FEs.
        self._max_fes: Final[Optional[int]] = check_max_fes(max_fes, True)
        #: A version of :attr:`_max_fes` that can be used in comparisons.
        self._end_fes: Final[int] = 9_223_372_036_854_775_800 \
            if (self._max_fes is None) else self._max_fes
        #: The goal objective value.
        self._goal_f: Final[Union[int, float, None]] = \
            check_goal_f(goal_f, True)
        #: A comparable version of :attr:`self._goal_f`.
        self._end_f: Union[int, float] = \
            -inf if (self._goal_f is None) else self._goal_f
        #: The currently consumed nanoseconds.
        self._current_time_nanos: int = 0
        #: The currently consumed objective function evaluations (FEs).
        self._current_fes: int = 0
        #: The time (in nanoseconds) when the last improvement was made.
        self._last_improvement_time_nanos: int = -1
        #: The FE when the last improvement was made.
        self._last_improvement_fe: int = -1
        #: The maximum runtime in milliseconds.
        self._max_time_millis: Final[Optional[int]] = \
            check_max_time_millis(max_time_millis, True)
        #: A comparable version of :attr:`_max_time_millis`, but representing
        #: the end time in nanoseconds rounded to the next highest
        #: millisecond.
        self._end_time_nanos: Final[Union[float, int]] = \
            inf if (self._max_time_millis is None) else \
            _ns_to_ms(int(self._start_time_nanos
                          + (1_000_000 * self._max_time_millis))) \
            * 1_000_000
        #: The timer until the end-of-run, or `None` if there is no end time.
        self.__timer: Final[Optional[Timer]] = None \
            if (self._max_time_millis is None) else \
            Timer(interval=self._max_time_millis / 1_000.0,
                  function=self.terminate)

        #: an internal base exception caught by the algorithm execution
        self._caught: Optional[BaseException] = None

    def _after_init(self) -> None:
        """
        Finish initialization, start timer for termination if needed.

        Internal method that must be called after __init__ is completed.
        """
        if self.__timer is not None:
            self.__timer.start()

    def should_terminate(self) -> bool:
        if self._terminated:
            self._knows_that_terminated = True
            return True
        return False

    def get_consumed_fes(self) -> int:
        return self._current_fes

    def get_consumed_time_millis(self) -> int:
        if not self._terminated:
            self._current_time_nanos = time = _TIME_IN_NS()
            if time >= self._end_time_nanos:
                self.terminate()
        return _ns_to_ms(self._current_time_nanos - self._start_time_nanos)

    def get_max_time_millis(self) -> Optional[int]:
        return self._max_time_millis

    def get_max_fes(self) -> Optional[int]:
        return self._max_fes

    def get_last_improvement_fe(self) -> int:
        if self._last_improvement_fe < 0:
            raise ValueError("Did not perform FE yet, cannot query "
                             "last improvement FE.")
        return self._last_improvement_fe

    def get_last_improvement_time_millis(self) -> int:
        if self._last_improvement_time_nanos < 0:
            raise ValueError("Did not perform FE yet, cannot query "
                             "last improvement time.")
        return _ns_to_ms(self._last_improvement_time_nanos
                         - self._start_time_nanos)

    def terminate(self) -> None:
        with self.__lock:
            old_terminated = self._terminated
            self._terminated = True
            if old_terminated:
                return
            if self.__timer is not None:
                self.__timer.cancel()
            del self.__timer
            self._current_time_nanos = _TIME_IN_NS()

    def get_copy_of_best_y(self, y) -> None:
        """
        Get a copy of the current best point in the solution space.

        This method in this internal class just forwards to
        :meth:`get_copy_of_best_x`.

        :param y: the destination data structure to be overwritten
        """
        return self.get_copy_of_best_x(y)

    def _log_own_parameters(self, logger: KeyValueLogSection) -> None:
        """
        Write the parameters of this process to the logger.

        This includes the limits on runtime and FEs.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        if not (self._max_fes is None):
            logger.key_value(KEY_MAX_FES, self._max_fes)
        if not (self._max_time_millis is None):
            logger.key_value(KEY_MAX_TIME_MILLIS,
                             self._max_time_millis)
        if not (self._goal_f is None):
            logger.key_value(KEY_GOAL_F, self._goal_f)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Write the standard parameters of this process to the logger.

        This includes the limits on runtime and FEs.

        :param logger: the logger
        """
        with logger.scope(SCOPE_PROCESS) as sc:
            self._log_own_parameters(sc)

    def __str__(self) -> str:
        """
        Get the name of this process implementation.

        :return: "baseProcess"
        """
        return "baseProcess"

    def add_log_section(self, title: str, text: str) -> None:
        """
        Add a section to the log, if a log is written (otherwise ignore it).

        :param title: the title of the log section
        :param text: the text to log
        """
        if not isinstance(title, str):
            raise type_error(title, "title", str)
        t = title.strip()
        if (len(t) != len(title)) or (len(t) <= 0) or (" " in t) \
                or ("\n" in t) or ("\t" in t):
            raise ValueError("Section title must not be empty or contain "
                             f"white space, but '{title}' is/does.")
        if (t in _ALL_SECTIONS) or (SECTION_START in t) or (SECTION_END in t):
            raise ValueError(f"title '{t}' is a forbidden section title")
        if t.upper() != t:
            raise ValueError("section titles must be in upper case,"
                             f"but yours is '{t}' (vs. '{t.upper()}'.")
        if not isinstance(text, str):
            raise type_error(text, "text", str)
        if (SECTION_START in t) or (SECTION_END in t):
            raise ValueError(
                f"text of section '{t}' must not contain '{SECTION_START}' or"
                f" '{SECTION_END}' but is '{text}'")



"""The algorithm execution API."""
from math import isfinite
from typing import Optional
from typing import Union

from moptipy.api._process_base import _ProcessBase
from moptipy.api._process_no_ss import _ProcessNoSS
from moptipy.api._process_no_ss_log import _ProcessNoSSLog
from moptipy.api._process_ss import _ProcessSS
from moptipy.api._process_ss_log import _ProcessSSLog
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import Process
from moptipy.api.process import check_max_fes, check_max_time_millis, \
    check_goal_f
from moptipy.api.space import Space, check_space
from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.path import Path


def _check_log_file(log_file: Optional[str],
                    none_is_ok: bool = True) -> Optional[Path]:
    """
    Check a log file.

    :param log_file: the log file
    :param none_is_ok: is `None` ok for log files?
    :return: the log file
    """
    if log_file is None:
        if none_is_ok:
            return None
    return Path.path(log_file)


class Execution:
    """
    Define all the components of an experiment and then execute it.

    This class follows the builder pattern. It allows us to
    step-by-step store all the parameters needed to execute an
    experiment. Via the method :meth:`~Execution.execute`, we can then
    run the experiment and obtain the instance of
    :class:`~moptipy.api.process.Process` *after* the execution of the
    algorithm. From this instance, we can query the final result of the
    algorithm application.
    """

    def __init__(self) -> None:
        """Initialize the execution builder."""
        super().__init__()
        self.__algorithm: Optional[Algorithm] = None
        self.__solution_space: Optional[Space] = None
        self.__objective: Optional[Objective] = None
        self.__search_space: Optional[Space] = None
        self.__encoding: Optional[Encoding] = None
        self.__rand_seed: Optional[int] = None
        self.__max_fes: Optional[int] = None
        self.__max_time_millis: Optional[int] = None
        self.__goal_f: Union[None, int, float] = None
        self.__log_file: Optional[Path] = None
        self.__log_improvements: bool = False
        self.__log_all_fes: bool = False

    def set_algorithm(self, algorithm: Algorithm) -> None:
        """
        Set the algorithm to be used for this experiment.

        :param algorithm: the algorithm
        """
        self.__algorithm = check_algorithm(algorithm)

    def get_algorithm(self) -> Algorithm:
        """
        Obtain the algorithm.

        Requires that :meth:`set_algorithm` was called first.

        :return: the algorithm
        """
        return check_algorithm(self.__algorithm)

    def set_solution_space(self, solution_space: Space) -> None:
        """
        Set the solution space to be used for this experiment.

        This is the space managing the data structure holding the candidate
        solutions.

        :param solution_space: the solution space
        """
        self.__solution_space = check_space(solution_space)

    def set_objective(self, objective: Objective) -> None:
        """
        Set the objective function to be used for this experiment.

        This is the function rating the quality of candidate solutions.

        :param objective: the objective function
        """
        self.__objective = check_objective(objective)

    def set_search_space(self, search_space: Optional[Space]) -> None:
        """
        Set the search space to be used for this experiment.

        This is the space from which the algorithm samples points.

        :param search_space: the search space, or `None` of none shall be
            used, i.e., if search and solution space are the same
        """
        self.__search_space = check_space(search_space, none_is_ok=True)

    def set_encoding(self, encoding: Optional[Encoding]) -> None:
        """
        Set the encoding to be used for this experiment.

        This is the function translating from the search space to the
        solution space.

        :param encoding: the encoding, or `None` of none shall be used
        """
        self.__encoding = check_encoding(encoding, none_is_ok=True)

    def set_rand_seed(self, rand_seed: Optional[int]) -> None:
        """
        Set the seed to be used for initializing the random number generator.

        :param rand_seed: the random seed, or `None` if a seed should
            automatically be chosen when the experiment is executed
        """
        self.__rand_seed = None if rand_seed is None \
            else rand_seed_check(rand_seed)

    def set_max_fes(self, max_fes: int,  # +book
                    force_override: bool = False) -> None:
        """
        Set the maximum FEs.

        This is the number of candidate solutions an optimization is allowed
        to evaluate. If this method is called multiple times, then the
        shortest limit is used unless `force_override` is `True`.

        :param max_fes: the maximum FEs
        :param force_override: the use the value given in `max_time_millis`
            regardless of what was specified before
        """
        max_fes = check_max_fes(max_fes)
        if not (self.__max_fes is None):
            if max_fes >= self.__max_fes:
                if not force_override:
                    return
        self.__max_fes = max_fes

    def set_max_time_millis(self, max_time_millis: int,
                            force_override: bool = False) -> None:
        """
        Set the maximum time in milliseconds.

        This is the maximum time that the process is allowed to run. If this
        method is called multiple times, the shortest time is used unless
        `force_override` is `True`.

        :param max_time_millis: the maximum time in milliseconds
        :param force_override: the use the value given in `max_time_millis`
            regardless of what was specified before
        """
        max_time_millis = check_max_time_millis(max_time_millis)
        if not (self.__max_time_millis is None):
            if max_time_millis >= self.__max_time_millis:
                if not force_override:
                    return
        self.__max_time_millis = max_time_millis

    def set_goal_f(self, goal_f: Union[int, float]) -> None:
        """
        Set the goal objective value after which the process can stop.

        If this method is called multiple times, then the largest value is
        retained.

        :param goal_f: the goal objective value.
        """
        goal_f = check_goal_f(goal_f)
        if not (self.__goal_f is None):
            if goal_f <= self.__goal_f:
                return
        self.__goal_f = goal_f

    def set_log_file(self, log_file: Optional[str]) -> None:
        """
        Set the log file to write to.

        This method can be called arbitrarily often.

        :param log_file: the log file
        """
        self.__log_file = _check_log_file(log_file, True)

    def set_log_improvements(self, log_improvements: bool = True) -> None:
        """
        Set whether improvements should be logged.

        :param log_improvements: if improvements should be logged?
        """
        if not isinstance(log_improvements, bool):
            raise ValueError("log improvements must be bool, but is "
                             f"{type(log_improvements)}.")
        self.__log_improvements = log_improvements

    def set_log_all_fes(self, log_all_fes: bool = True) -> None:
        """
        Set whether all FEs should be logged.

        :param log_all_fes: if all FEs should be logged?
        """
        if not isinstance(log_all_fes, bool):
            raise ValueError(
                f"log_all_FEs must be bool, but is {type(log_all_fes)}.")
        self.__log_all_fes = log_all_fes

    def execute(self) -> Process:
        """
        Execute the experiment and return the process *after* the run.

        The optimization process constructed with this object is executed.
        This means that first, an instance of
        :class:`~moptipy.api.process.Process` is constructed.
        Then, the method :meth:`~moptipy.api.algorithm.Algorithm.solve` is
        applied to this instance.
        In other words, the optimization algorithm is executed until it
        terminates.
        Finally, this method returns the :class:`~moptipy.api.process.Process`
        instance *after* algorithm completion.
        This instance then can be queried for the final result of the run (via
        :meth:`~moptipy.api.process.Process.get_copy_of_best_y`), the
        objective value of this final best solution (via
        :meth:`~moptipy.api.process.Process.get_best_f`), and other
        information.

        :return: the process *after* the run, i.e., in the state where it can
            be queried for the result
        """
        algorithm = check_algorithm(self.__algorithm)
        solution_space = check_space(self.__solution_space)
        objective = check_objective(self.__objective)
        search_space = check_space(self.__search_space,
                                   self.__encoding is None)
        encoding = check_encoding(self.__encoding,
                                  search_space is None)
        rand_seed = self.__rand_seed
        if not (rand_seed is None):
            rand_seed = rand_seed_check(rand_seed)
        max_time_millis = check_max_time_millis(self.__max_time_millis, True)
        max_fes = check_max_fes(self.__max_fes, True)
        goal_f = check_goal_f(self.__goal_f, True)
        f_lb = objective.lower_bound()
        if (not (f_lb is None)) and isfinite(f_lb):
            if (goal_f is None) or (f_lb > goal_f):
                goal_f = f_lb

        log_all_fes = self.__log_all_fes
        log_improvements = self.__log_improvements or self.__log_all_fes

        log_file = self.__log_file
        if log_file is None:
            if log_all_fes:
                raise ValueError("Log file cannot be None "
                                 "if all FEs should be logged.")
            if log_improvements:
                raise ValueError("Log file cannot be None "
                                 "if improvements should be logged.")

        else:
            log_file.create_file_or_truncate()

        process: _ProcessBase
        if search_space is None:
            if log_improvements or log_all_fes:
                process = _ProcessNoSSLog(solution_space=solution_space,
                                          objective=objective,
                                          algorithm=algorithm,
                                          log_file=log_file,
                                          rand_seed=rand_seed,
                                          max_fes=max_fes,
                                          max_time_millis=max_time_millis,
                                          goal_f=goal_f,
                                          log_all_fes=log_all_fes)
            else:
                process = _ProcessNoSS(solution_space=solution_space,
                                       objective=objective,
                                       algorithm=algorithm,
                                       log_file=log_file,
                                       rand_seed=rand_seed,
                                       max_fes=max_fes,
                                       max_time_millis=max_time_millis,
                                       goal_f=goal_f)
        else:
            if log_improvements or log_all_fes:
                process = _ProcessSSLog(
                    solution_space=solution_space,
                    objective=objective,
                    algorithm=algorithm,
                    search_space=search_space,
                    encoding=encoding,
                    log_file=log_file,
                    rand_seed=rand_seed,
                    max_fes=max_fes,
                    max_time_millis=max_time_millis,
                    goal_f=goal_f,
                    log_all_fes=log_all_fes)
            else:
                process = _ProcessSS(solution_space=solution_space,
                                     objective=objective,
                                     algorithm=algorithm,
                                     search_space=search_space,
                                     encoding=encoding,
                                     log_file=log_file,
                                     rand_seed=rand_seed,
                                     max_fes=max_fes,
                                     max_time_millis=max_time_millis,
                                     goal_f=goal_f)
        try:
            # noinspection PyProtectedMember
            process._after_init()
            algorithm.solve(process)
        except BaseException as be:
            # noinspection PyProtectedMember
            process._caught = be
        return process



"""The base class for spaces based on numpy arrays."""
from typing import Final

import numpy

from moptipy.api.logging import KEY_SPACE_NUM_VARS
from moptipy.api.space import Space
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import KEY_NUMPY_TYPE
from moptipy.utils.types import type_error


class NPArraySpace(Space):
    """
    A space where each element is a one-dimensional numpy array.

    Such spaces can serve as basis for implementing combinatorial
    optimization and can be extended to host permutations.
    """

    def __init__(self, dimension: int, dtype: numpy.dtype) -> None:
        """
        Create the numpy array-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param dtype: the data type
        """
        if not isinstance(dimension, int):
            raise type_error(dimension, "dimension", int)
        if (dimension < 1) or (dimension > 1_000_000_000):
            raise ValueError("dimension must be in 1..1_000_000_000, "
                             f"but got {dimension}.")

        if not isinstance(dtype, numpy.dtype):
            raise type_error(dtype, "dtype", numpy.dtype)
        if (not isinstance(dtype.char, str)) or (len(dtype.char) != 1):
            raise ValueError(
                f"dtype.char must be str of length 1, but is {dtype.char}")

        #: The basic data type of the vector space elements.
        self.dtype: Final[numpy.dtype] = dtype
        #: The dimension, i.e., the number of elements of the vectors.
        self.dimension: Final[int] = dimension
        # the function forwards
        self.copy = numpy.copyto  # type: ignore
        self.is_equal = numpy.array_equal  # type: ignore

    def create(self) -> numpy.ndarray:
        """
        Create a vector with all zeros.

        :return: the vector
        """
        return numpy.zeros(shape=self.dimension, dtype=self.dtype)

    def to_str(self, x: numpy.ndarray) -> str:
        """
        Convert an integer string to a string, using `,` as separator.

        :param x: the integer string
        :return: the string
        """
        return ",".join([str(xx) for xx in x])

    def from_str(self, text: str) -> numpy.ndarray:
        """
        Convert a string to an integer string.

        :param text: the text
        :return: the vector
        :raises TypeError: if `text` is not a `str`
        :raises ValueError: if `text` cannot be converted to a valid vector
        """
        if not isinstance(text, str):
            raise type_error(text, "text", str)
        x = numpy.fromstring(text, dtype=self.dtype, sep=",")
        self.validate(x)
        return x

    def validate(self, x: numpy.ndarray) -> None:
        """
        Validate a string.

        :param x: the integer string
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        if not isinstance(x, numpy.ndarray):
            raise type_error(x, "x", numpy.ndarray)
        if x.dtype != self.dtype:
            raise ValueError(
                f"x must be of type {self.dtype} but is of type {x.dtype}.")
        if (len(x.shape) != 1) or (x.shape[0] != self.dimension):
            raise ValueError(f"x must be of shape ({self.dimension}) but is "
                             f"of shape {x.shape}.")

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_SPACE_NUM_VARS, self.dimension)
        logger.key_value(KEY_NUMPY_TYPE, self.dtype.char)



"""
The implementation of the basic hill climbing algorithm `hc`.

The algorithm starts by applying the nullary search operator, an
implementation of :meth:`~moptipy.api.operators.Op0.op0`, to sample one fully
random solution. This is the first best-so-far solution. In each step, it
applies the unary operator, an implementation of
:meth:`~moptipy.api.operators.Op1.op1`, to the best-so-far solution to obtain
a new, similar solution. If this new solution is strictly better than the
current best-so-far solution, it replaces this solution. Otherwise, it is
discarded.

The hill climbing algorithm is a simple local search that only accepts
strictly improving moves. It is thus similar to the randomized local search
(`rls`) implemented in :class:`~moptipy.algorithms.rls.RLS`, which, however,
accepts non-deteriorating moves. We also provide `hcr`, a variant of the hill
climber that restarts automatically with a certain number of moves were not
able to improve the current best-so-far solution in class :class:`~moptipy.\
algorithms.hill_climber_with_restarts.HillClimberWithRestarts`.
"""
from typing import Final, Union, Callable

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error


# start book
class HillClimber(Algorithm1):
    """
    The stochastic hill climbing algorithm only accepts improving moves.

    In each step, a hill climber creates a modified copy `new_x` of the
    current best solution `best_x`. If `new_x` is better than `best_x`,
    it becomes the new `best_x`. Otherwise, it is discarded.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the hill climber to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, best_x)  # Create 1 solution randomly and
        best_f: Union[int, float] = evaluate(best_x)  # evaluate it.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            new_f: Union[int, float] = evaluate(new_x)
            if new_f < best_f:  # new_x is _better_ than best_x?
                best_f = new_f  # Store its objective value.
                best_x, new_x = new_x, best_x  # Swap best and new.
# end book

    def __solve_seeded(self, process: Process) -> None:
        """
        Apply the hill climber with a seed to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at an existing point in the search space and get its quality.
        process.get_copy_of_best_x(best_x)  # get the best-so-far solution
        best_f: Union[int, float] = process.get_best_f()  # get the quality.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            new_f: Union[int, float] = evaluate(new_x)
            if new_f < best_f:  # new_x is _better_ than best_x?
                best_f = new_f  # Store its objective value.
                best_x, new_x = new_x, best_x  # Swap best and new.

    def __init__(self, op0: Op0, op1: Op1,
                 seeded: bool = False) -> None:
        """
        Create the hill climber.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param seeded: `True` if the algorithm should be run in a seeded
            fashion, i.e., expect an existing best solution. `False` if
            it should run in the traditional way, starting at a random
            solution
        """
        super().__init__("hc", op0, op1)
        if not isinstance(seeded, bool):
            raise type_error(seeded, "seeded", bool)
        if seeded:
            self.solve = self.__solve_seeded  # type: ignore
        #: was this algorithm started in its seeded fashion?
        self.__seeded: Final[bool] = seeded

    def log_parameters_to(self, logger: KeyValueLogSection):
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("seeded", self.__seeded)



"""This module provides the base class for implementing encodings."""
from typing import Optional

from moptipy.api.component import Component
from moptipy.utils.types import type_error


# start book
class Encoding(Component):
    """The encodings translates from a search space to a solution space."""

    def map(self, x, y) -> None:
        """
        Translate from search- to solution space.

        Map a point `x` from the search space to a point `y`
        in the solution space.

        :param x: the point in the search space, remaining unchanged.
        :param y: the destination data structure for the point in the
            solution space, whose contents will be overwritten
        """
    # end book


def check_encoding(encoding: Optional[Encoding],
                   none_is_ok: bool = True) -> Optional[Encoding]:
    """
    Check whether an object is a valid instance of :class:`Encoding`.

    :param encoding: the object
    :param none_is_ok: is it ok if `None` is passed in?
    :return: the object
    :raises TypeError: if `encoding` is not an instance of :class:`Encoding`
    """
    if not isinstance(encoding, Encoding):
        if none_is_ok and (encoding is None):
            return None
        raise type_error(encoding, "encoding", Encoding)
    return encoding



"""An implementation of processes with different search and solution spaces."""
from typing import Optional, Union, Final, Callable

from moptipy.api import logging
from moptipy.api._process_base import _TIME_IN_NS
from moptipy.api._process_no_ss import _ProcessNoSS
from moptipy.api.algorithm import Algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.objective import Objective
from moptipy.api.space import Space, check_space
from moptipy.utils.logger import KeyValueLogSection, Logger
from moptipy.utils.path import Path


class _ProcessSS(_ProcessNoSS):
    """A class implementing a process with search and solution space."""

    def __init__(self,
                 solution_space: Space,
                 objective: Objective,
                 algorithm: Algorithm,
                 log_file: Optional[Path] = None,
                 search_space: Space = None,
                 encoding: Encoding = None,
                 rand_seed: Optional[int] = None,
                 max_fes: Optional[int] = None,
                 max_time_millis: Optional[int] = None,
                 goal_f: Union[int, float, None] = None) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param search_space: the search space.
        :param encoding: the encoding
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the
            process is terminated
        """
        super().__init__(solution_space=solution_space,
                         objective=objective,
                         algorithm=algorithm,
                         log_file=log_file,
                         rand_seed=rand_seed,
                         max_fes=max_fes,
                         max_time_millis=max_time_millis,
                         goal_f=goal_f)

        #: The search space.
        self._search_space: Final[Space] = check_space(search_space)
        #: The encoding.
        self._encoding: Final[Encoding] = check_encoding(encoding)
        #: the internal encoder
        self._g: Final[Callable] = self._encoding.map
        #: The holder for the currently de-coded solution.
        self._current_y = self._solution_space.create()
        #: The current best point in the search space.
        self._current_best_x: Final = self._search_space.create()
        # wrappers
        self.create = self._search_space.create  # type: ignore
        self.copy = self._search_space.copy  # type: ignore
        self.to_str = self._search_space.to_str  # type: ignore
        self.is_equal = self._search_space.is_equal  # type: ignore
        self.from_str = self._search_space.from_str  # type: ignore
        self.n_points = self._search_space.n_points  # type: ignore
        self.validate = self._search_space.validate  # type: ignore

    def evaluate(self, x) -> Union[float, int]:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError('The process has been terminated and the '
                                 'algorithm knows it.')
            return self._current_best_f

        current_y: Final = self._current_y
        self._g(x, current_y)
        result: Final[Union[int, float]] = self._f(current_y)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        if result < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = result
            self.copy(self._current_best_x, x)
            self._current_y = self._current_best_y
            self._current_best_y = current_y
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (result <= self._end_f)

        if do_term:
            self.terminate()

        return result

    def register(self, x, f: Union[int, float]) -> None:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError('The process has been terminated and the '
                                 'algorithm knows it.')
            return

        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        if f < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = f
            self.copy(self._current_best_x, x)
            current_y: Final = self._current_y
            self._g(x, current_y)
            self._current_y = self._current_best_y
            self._current_best_y = current_y
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (f <= self._end_f)

        if do_term:
            self.terminate()

    def get_copy_of_best_x(self, x) -> None:
        if self._current_fes > 0:
            return self.copy(x, self._current_best_x)
        raise ValueError('No current best x available.')

    def get_copy_of_best_y(self, y):
        if self._current_fes > 0:
            return self._copy_y(y, self._current_best_y)
        raise ValueError('No current best y available.')

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        super().log_parameters_to(logger)
        with logger.scope(logging.SCOPE_SEARCH_SPACE) as sc:
            self._search_space.log_parameters_to(sc)
        with logger.scope(logging.SCOPE_ENCODING) as sc:
            self._encoding.log_parameters_to(sc)

    def _write_log(self, logger: Logger) -> None:
        # noinspection PyProtectedMember
        super()._write_log(logger)

        if self._current_fes > 0:
            with logger.text(logging.SECTION_RESULT_X) as txt:
                txt.write(self._search_space.to_str(self._current_best_x))

    def _validate_x(self) -> None:
        """Validate x, if it exists."""
        self._search_space.validate(self._current_best_x)

    def __str__(self) -> str:
        return "ProcessWithSearchSpace"



"""
An operator swapping two elements in a permutation.

This is a unary search operator which first copies the source string `x` to
the destination string `dest`. Then it draws an index `i1` randomly.
It keeps drawing a second random index `i2` until `dest[i1] != dest[i2]`,
i.e., until the elements at the two indices are different. This will always
be true for actual permutations if `i1 != i2`, but for permutations with
repetitions, even if `i1 != i2`, sometimes `dest[i1] == dest[i2]`. Anyway,
as soon as the elements at `i1` and `i2` are different, they will be swapped.

This operator performs one swap. It is similar to :class:`~moptipy.operators.\
permutations.op1_swapn.Op1SwapN`, which performs a random number of swaps.
"""
from typing import Final, Callable

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1


# start book
class Op1Swap2(Op1):
    """
    This unary search operation swaps two (different) elements.

    In other words, it performs exactly one swap on a permutation.
    It spans a neighborhood of a rather limited size but is easy
    and fast.
    """

    def op1(self, random: Generator,
            dest: np.ndarray, x: np.ndarray) -> None:
        """
        Copy `x` into `dest` and swap two different values in `dest`.

        :param random: the random number generator
        :param dest: the array to receive the modified copy of `x`
        :param x: the existing point in the search space
        """
        np.copyto(dest, x)  # First, we copy `x` to `dest`.
        length: Final[int] = len(dest)  # Get the length of `dest`.
        ri: Final[Callable[[int], int]] = random.integers  # fast call!

        i1: Final[int] = ri(length)  # Get the first random index.
        v1: Final = dest[i1]  # Get the value at the first index.
        while True:  # Repeat until we find a different value.
            i2: int = ri(length)  # Get the second random index.
            v2 = dest[i2]  # Get the value at the second index.
            if v1 != v2:  # If both values different...
                dest[i2] = v1  # store v1 where v2 was
                dest[i1] = v2  # store v2 where v1 was
                return  # Exit function: we are finished.
    # end book

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "swap2", the name of this operator
        :retval "swap2": always
        """
        return "swap2"



"""Functions that can be used to test search operators."""
from math import isqrt
from typing import Optional, Callable, Any, Union

from numpy.random import default_rng

from moptipy.api.operators import Op0
from moptipy.api.space import Space
from moptipy.tests.component import validate_component


def validate_op0(op0: Op0,
                 search_space: Space = None,
                 make_search_space_element_valid:
                 Optional[Callable[[Any], Any]] = lambda x: x,
                 number_of_samples: int = 100,
                 min_unique_samples: Union[int, Callable[[int, Space], int]]
                 = lambda samples, space:
                 max(1, min(samples // 2, isqrt(space.n_points())))) -> None:
    """
    Check whether an object is a moptipy nullary operator.

    :param op0: the operator
    :param search_space: the search space
    :param make_search_space_element_valid: make a point in the search
        space valid
    :param number_of_samples: the number of times to invoke the operator
    :param min_unique_samples: a lambda for computing the number
    :raises ValueError: if `op0` is not a valid `Op0Shuffle`
    """
    if not isinstance(op0, Op0):
        raise ValueError("Expected to receive an instance of Op0Shuffle, but "
                         f"got a {type(op0)}.")
    validate_component(op0)

    count: int = 0
    if search_space is not None:
        count += 1
    if make_search_space_element_valid is not None:
        count += 1
    if count <= 0:
        return
    if count < 2:
        raise ValueError(
            "either provide both of search_space and "
            "make_search_space_element_valid or none.")

    x = search_space.create()
    if x is None:
        raise ValueError("Space must not return None.")

    x = make_search_space_element_valid(x)
    if x is None:
        raise ValueError("Make valid failed.")
    search_space.validate(x)

    random = default_rng()
    seen = set()

    if not (hasattr(op0, 'op0') and callable(getattr(op0, 'op0'))):
        raise ValueError("op0 must have method op0.")
    for _ in range(number_of_samples):
        op0.op0(random, x)
        search_space.validate(x)
        strstr = search_space.to_str(x)
        if (not isinstance(strstr, str)) or (len(strstr) <= 0):
            raise ValueError("to_str produces either no string or "
                             f"empty string, namely '{strstr}'.")
        seen.add(strstr)

    expected: int
    if callable(min_unique_samples):
        expected = min_unique_samples(number_of_samples, search_space)
    else:
        expected = min_unique_samples
    if not isinstance(expected, int):
        raise ValueError(f"expected number of unique values must be int,"
                         f" but is {type(expected)}.")
    if expected <= 0:
        raise ValueError(f"expected number of unique values must be positive,"
                         f" but is {expected}.")
    if len(seen) < expected:
        raise ValueError(
            f"It is expected that at least {expected} different elements "
            "will be created by nullary search operator from "
            f"{number_of_samples} samples, but we only got {len(seen)} "
            "different points.")



"""Providing a process without explicit logging with a single space."""
import os
from io import StringIO
from math import inf, isfinite
from traceback import print_tb
from typing import Optional, Union, Final, Callable, Dict

from numpy.random import Generator

from moptipy.api import logging
from moptipy.api._process_base import _ProcessBase, _ns_to_ms, _TIME_IN_NS
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import check_goal_f
from moptipy.api.space import Space, check_space
from moptipy.utils.logger import KeyValueLogSection, FileLogger
from moptipy.utils.logger import Logger
from moptipy.utils.nputils import rand_generator, rand_seed_generate, \
    rand_seed_check
from moptipy.utils.path import Path
from moptipy.utils.sys_info import log_sys_info
from moptipy.utils.types import type_name_of, type_error


def _error_1(logger: Logger, title: str, exception_type,
             exception_value, traceback):
    """
    Create a text section with error information as from a contextmanager.

    :param logger: the logger to write to
    :param title: the title of the section with error information to be
        created
    :param exception_type: the exception type
    :param exception_value: the exception value
    :param traceback: the traceback
    """
    if exception_type or exception_value or traceback:
        with logger.text(title=title) as ts:
            if exception_type:
                ts.write(logging.KEY_EXCEPTION_TYPE)
                ts.write(": ")
                if isinstance(exception_type, str):
                    if exception_type.startswith("<class '"):
                        exception_type = exception_type[8:-2]
                    ts.write(exception_type.strip())
                else:
                    ts.write(type_name_of(exception_type))
                ts.write(os.linesep)
            if exception_value:
                ts.write(logging.KEY_EXCEPTION_VALUE)
                ts.write(": ")
                ts.write(exception_value.strip())
                ts.write(os.linesep)
            if traceback:
                ts.write(logging.KEY_EXCEPTION_STACK_TRACE)
                ts.write(":")
                ts.write(os.linesep)
                sio = StringIO()
                print_tb(traceback, file=sio)
                for line in sio.getvalue().split("\n"):
                    ts.write(line.strip())
                    ts.write(os.linesep)


def _error_2(logger: Logger, title: str, exception: BaseException):
    """
    Log an exception.

    :param logger: the logger to write to
    :param title: the title of the section with error information to be
        created
    :param exception: the exception

    >>> from moptipy.utils.logger import InMemoryLogger
    >>> ime = InMemoryLogger()
    >>> def k():
    ...     1 / 0
    >>> try:
    ...     k()
    ... except BaseException as be:
    ...     _error_2(ime, "ERROR", be)
    >>> print(ime.get_log()[0])
    BEGIN_ERROR
    >>> print(ime.get_log()[1])
    exceptionType: ZeroDivisionError
    >>> print(ime.get_log()[2])
    exceptionValue: division by zero
    >>> print(ime.get_log()[3])
    exceptionStackTrace:
    >>> print(ime.get_log()[-1])
    END_ERROR
    """
    _error_1(logger, title, exception_type=exception,
             exception_value=str(exception),
             traceback=exception.__traceback__)


class _ProcessNoSS(_ProcessBase):
    """
    An internal class process implementation.

    This class implements a stand-alone process without explicit logging where
    the search and solution space are the same.
    """

    def __init__(self,
                 solution_space: Space,
                 objective: Objective,
                 algorithm: Algorithm,
                 log_file: Optional[Path] = None,
                 rand_seed: Optional[int] = None,
                 max_fes: Optional[int] = None,
                 max_time_millis: Optional[int] = None,
                 goal_f: Union[int, float, None] = None) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the search- and solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the process
            is terminated
        """
        super().__init__(max_fes=max_fes,
                         max_time_millis=max_time_millis,
                         goal_f=goal_f)

        #: The solution space, i.e., the data structure of possible solutions.
        self._solution_space: Final[Space] = check_space(solution_space)
        #: The objective function rating candidate solutions.
        self.__objective: Final[Objective] = check_objective(objective)
        #: the internal invoker for the objective function
        self._f: Final[Callable] = self.__objective.evaluate
        #: The algorithm to be applied.
        self.__algorithm: Final[Algorithm] = check_algorithm(algorithm)
        #: The random seed.
        self.__rand_seed: Final[int] = rand_seed_generate() \
            if rand_seed is None \
            else rand_seed_check(rand_seed)
        #: The random number generator.
        self.__random: Final[Generator] = rand_generator(self.__rand_seed)
        #: The current best solution.
        self._current_best_y = self._solution_space.create()
        #: The current best objective value
        self._current_best_f: Union[int, float] = inf
        #: The log file, or `None` is needed
        if log_file is not None:
            if not isinstance(log_file, Path):
                raise type_error(log_file, "log_file", Path)
        self.__log_file: Final[Optional[Path]] = log_file
        #: the method for copying y
        self._copy_y: Final[Callable] = self._solution_space.copy
        #: set up the method forwards
        self.lower_bound = self.__objective.lower_bound  # type: ignore
        if self._end_f <= -inf:
            self._end_f = check_goal_f(self.lower_bound())
            self.lower_bound = lambda: self._end_f  # type: ignore
        self.upper_bound = self.__objective.upper_bound  # type: ignore
        self.create = self._solution_space.create  # type: ignore
        self.copy = self._solution_space.copy  # type: ignore
        self.to_str = self._solution_space.to_str  # type: ignore
        self.is_equal = self._solution_space.is_equal  # type: ignore
        self.from_str = self._solution_space.from_str  # type: ignore
        self.n_points = self._solution_space.n_points  # type: ignore
        self.validate = self._solution_space.validate  # type: ignore
        #: the internal section logger
        self.__sections: Optional[Dict[str, str]] = \
            None if log_file is None else {}

    def get_random(self) -> Generator:
        return self.__random

    def evaluate(self, x) -> Union[float, int]:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError('The process has been terminated and '
                                 'the algorithm knows it.')
            return self._current_best_f

        result: Final[Union[int, float]] = self._f(x)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        if result < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = result
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (result <= self._end_f)
            self._copy_y(self._current_best_y, x)

        if do_term:
            self.terminate()

        return result

    def register(self, x, f: Union[int, float]) -> None:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError('The process has been terminated and '
                                 'the algorithm knows it.')
            return

        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        if f < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = f
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (f <= self._end_f)
            self._copy_y(self._current_best_y, x)

        if do_term:
            self.terminate()

    def has_best(self) -> bool:
        return self._current_fes > 0

    def get_best_f(self) -> Union[int, float]:
        if self._current_fes > 0:
            return self._current_best_f
        raise ValueError('No current best available.')

    def get_copy_of_best_x(self, x) -> None:
        if self._current_fes > 0:
            return self._copy_y(x, self._current_best_y)
        raise ValueError('No current best available.')

    def _log_own_parameters(self, logger: KeyValueLogSection) -> None:
        super()._log_own_parameters(logger)
        logger.key_value(logging.KEY_RAND_SEED, self.__rand_seed, True)
        logger.key_value(logging.KEY_RAND_GENERATOR_TYPE,
                         type_name_of(self.__random))
        logger.key_value(logging.KEY_RAND_BIT_GENERATOR_TYPE,
                         type_name_of(self.__random.bit_generator))

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        super().log_parameters_to(logger)
        with logger.scope(logging.SCOPE_ALGORITHM) as sc:
            self.__algorithm.log_parameters_to(sc)
        with logger.scope(logging.SCOPE_SOLUTION_SPACE) as sc:
            self._solution_space.log_parameters_to(sc)
        with logger.scope(logging.SCOPE_OBJECTIVE_FUNCTION) as sc:
            self.__objective.log_parameters_to(sc)

    def add_log_section(self, title: str, text: str) -> None:
        super().add_log_section(title, text)
        if self.__sections is not None:
            if title in self.__sections:
                raise ValueError(f"section '{title}' already logged.")
            self.__sections[title] = text.strip()

    def _write_log(self, logger: Logger) -> None:
        """Write the information gathered during optimization into the log."""
        with logger.key_values(logging.SECTION_FINAL_STATE) as kv:
            kv.key_value(logging.KEY_TOTAL_FES, self._current_fes)
            kv.key_value(logging.KEY_TOTAL_TIME_MILLIS,
                         _ns_to_ms(self._current_time_nanos
                                   - self._start_time_nanos))
            if self._current_fes > 0:
                kv.key_value(logging.KEY_BEST_F, self._current_best_f)
                kv.key_value(logging.KEY_LAST_IMPROVEMENT_FE,
                             self._last_improvement_fe)
                kv.key_value(logging.KEY_LAST_IMPROVEMENT_TIME_MILLIS,
                             _ns_to_ms(self._last_improvement_time_nanos
                                       - self._start_time_nanos))

        with logger.key_values(logging.SECTION_SETUP) as kv:
            self.log_parameters_to(kv)

        log_sys_info(logger)

        if self._current_fes > 0:
            with logger.text(logging.SECTION_RESULT_Y) as txt:
                txt.write(self._solution_space.to_str(self._current_best_y))

    def _validate_x(self) -> None:
        """Validate x, if it exists."""

    def _check_timing(self) -> None:
        """
        Check whether there has been any timing errors.

        :raises ValueError: if there is any timing error
        """
        if self._current_time_nanos < self._start_time_nanos:
            raise ValueError(
                f"current_time_nanos={self._current_time_nanos} < "
                f"start_time_nanos={self._start_time_nanos}")
        if self._current_fes <= 0:
            raise ValueError("no FE was performed")
        if self._current_fes < self._last_improvement_fe:
            raise ValueError(
                f"current_fe={self._current_fes} < "
                f"last_improvement_fe={self._last_improvement_fe}")
        if self._current_time_nanos < self._last_improvement_time_nanos:
            raise ValueError(
                f"current_time_nanos={self._current_time_nanos} < "
                "last_improvement_time_nanos="
                f"{self._last_improvement_time_nanos}")

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """Exit the process and write the log if necessary."""
        # noinspection PyProtectedMember
        super().__exit__(exception_type, exception_value, traceback)

        # Update the total consumed time, but not include the error checks
        # below.
        self._current_time_nanos = _TIME_IN_NS()

        y_error: Optional[BaseException] = None  # error in solution?
        v_error: Optional[BaseException] = None  # error in objective value?
        x_error: Optional[BaseException] = None  # error in search space?
        t_error: Optional[BaseException] = None  # error in timing?
        log_error: Optional[BaseException] = None  # error while logging?
        try:
            self._solution_space.validate(self._current_best_y)
        except BaseException as be:
            y_error = be
        try:
            ff = self._f(self._current_best_y)
            if ff != self._current_best_f:
                raise ValueError(
                    "We re-computed the objective value of the best solution"
                    f"and got {ff}, but it has been registered as "
                    f"{self._current_best_f}!")
            if not isfinite(ff):
                raise ValueError(
                    f"Reproduced the objective value {ff} of the best "
                    "solution, but it is not finite?")
        except BaseException as be:
            v_error = be
        try:
            self._validate_x()
        except BaseException as be:
            x_error = be
        try:
            self._check_timing()
        except BaseException as be:
            t_error = be

        if self.__log_file is not None:
            with FileLogger(self.__log_file) as logger:
                try:
                    self._write_log(logger)
                except BaseException as be:
                    log_error = be

                if self._caught is not None:
                    _error_2(logger, logging.SECTION_ERROR_IN_RUN,
                             self._caught)
                if exception_type or exception_value or traceback:
                    _error_1(logger, logging.SECTION_ERROR_IN_CONTEXT,
                             exception_type, exception_value, traceback)
                if y_error:
                    _error_2(logger, logging.SECTION_ERROR_INVALID_Y, y_error)
                if v_error:
                    _error_2(logger, logging.SECTION_ERROR_BEST_F, v_error)
                if x_error:
                    _error_2(logger, logging.SECTION_ERROR_INVALID_X, x_error)
                if t_error:
                    _error_2(logger, logging.SECTION_ERROR_TIMING, t_error)
                if log_error:
                    _error_2(logger, logging.SECTION_ERROR_IN_LOG, log_error)

                # flush all the additional log sections at the end
                for t in sorted(self.__sections.keys()):
                    with logger.text(t) as sec:
                        sec.write(self.__sections[t])
                del self.__sections

        if not exception_type:
            # if no error happened when closing the process, raise any error
            # caught during validation.
            if self._caught is not None:
                raise self._caught  # pylint: disable=[E0702]
            if y_error:
                raise y_error
            if v_error:
                raise v_error
            if x_error:
                raise x_error
            if t_error:
                raise t_error
            if log_error:
                raise log_error

    def __str__(self) -> str:
        return "ProcessWithoutSearchSpace"
